from celery import Celery
from fastapi import FastAPI
from msdss_base_api import API
from msdss_base_database import Database
from msdss_data_api.managers import DataManager

from .env import *
from .managers import *
from .routers import *

class ModelsAPI(API):
    """
    Models API class for managing models.
    
    * Extends the :class:`msdss_base_api:msdss_base_api.core.API` class

    Parameters
    ----------
    users_api : :class:`msdss_users_api:msdss_users_api.core.UsersAPI` or None
        Users API object to enable user authentication for models routes.
        If ``None``, user authentication will not be used for models routes.
    database : :class:`msdss_base_database:msdss_base_database.core.Database` or None
        A database object for using models with data from the database. If ``None``, a default database object will be created.
    models : list(:class:`msdss_models_api.models.Model`) or dict(:class:`msdss_models_api.models.Model`)
        List or dict of available ``Model`` objects to use for creating and managing model instances.
        If ``list``, ensure that the class names are unique, otherwise the last object in the list takes priority.
    worker : :class:`celery:celery.Celery` or None
        Celery worker for background processes. If ``None``, a default worker will be created using parameters ``broker_url`` and ``backend_url``.
    broker_url : str or None
        Link to connect to a `Redis <https://redis.io/>`_ broker. Env vars will take priority - see parameter ``env``.
        If parameter ``worker`` is set, this will not be applied.
    backend_url : str or None
        Link to connect to a `Redis <https://redis.io/>`_ backend. Env vars will take priority - see parameter ``env``.
        If parameter ``worker`` is set, this will not be applied.
    folder : str
        The folder path to store models in. The folder will be created if it does not exist. Env vars will take priority - see parameter ``env``.
    models_router_settings : dict
        Keyword arguments passed to :func:`msdss_models_api.routers.get_models_router` except ``bg_manager``.
    load_env : bool
        Whether to load variables from a file with environmental variables at ``env_file`` or not.
    env : :class:`msdss_users_api.env.ModelsDotEnv`
        An object to set environment variables related to users configuration.
        These environment variables will overwrite the parameters above if they exist.

        By default, the related parameters above are assigned to each of the environment variables seen below if ``load_env`` is ``True``:

        .. jupyter-execute::
            :hide-code:

            from msdss_models_api.defaults import DEFAULT_DOTENV_KWARGS
            defaults = {k:v for k, v in DEFAULT_DOTENV_KWARGS.items() if k not in ['defaults', 'env_file', 'key_path']}
            print('<parameter> = <environment variable>\\n')
            for k, v in defaults.items():
                print(k + ' = ' + v)

    api : :class:`fastapi:fastapi.FastAPI`
        API object for creating routes.
    *args, **kwargs
        Additional arguments passed to :class:`msdss_base_api:msdss_base_api.core.API`.

    Attributes
    ----------
    models_api_database : :class:`msdss_base_database:msdss_base_database.core.Database`
        Database object for users API.
    models_api_worker : :class:`celery:celery.Celery`
        Same as parameter ``worker``.
    misc : dict
        Dictionary of miscellaneous values:

        * ``fastapi_users_objects`` (dict): dict of values returned from :func:`msdss_users_api.tools.create_fastapi_users_objects`

    Author
    ------
    Richard Wen <rrwen.dev@gmail.com>

    Example
    -------

    Create models api without users:

    .. jupyter-execute::

        from msdss_base_database import Database
        from msdss_models_api.models import Model
        from msdss_models_api import ModelsAPI

        # Create database object
        database = Database(
            driver='postgresql',
            user='msdss',
            password='msdss123',
            host='localhost',
            port='5432',
            database='msdss'
        )

        # Create models api without users
        app = ModelsAPI(
            models=[Model],
            database=database,
            broker_url='redis://localhost:6379/0',
            backend_url='redis://localhost:6379/0'
        )

        # Run the app with app.start()
        # Try API at http://localhost:8000/docs
        # app.start()

    Create Models API with users:

    .. jupyter-execute::

        from msdss_base_database import Database
        from msdss_models_api.models import Model
        from msdss_models_api import ModelsAPI
        from msdss_users_api import UsersAPI

        # Create database object
        database = Database(
            driver='postgresql',
            user='msdss',
            password='msdss123',
            host='localhost',
            port='5432',
            database='msdss'
        )

        # Create a users api
        # CHANGE SECRETS TO STRONG PHRASES
        users_api = UsersAPI(
            'cookie-secret',
            'jwt-secret',
            'reset-secret',
            'verification-secret',
            database=database
        )

        # Create a models api with users
        app = ModelsAPI(
            users_api,
            models=[Model],
            database=database,
            broker_url='redis://localhost:6379/0',
            backend_url='redis://localhost:6379/0'
        )

        # Add users routes
        app.add_app(users_api)

        # Run the app with app.start()
        # Try API at http://localhost:8000/docs
        # app.start()

    Create Models API with users and data management:

    .. jupyter-execute::

        from msdss_base_database import Database
        from msdss_models_api.models import Model
        from msdss_models_api import ModelsAPI
        from msdss_users_api import UsersAPI
        from msdss_data_api import DataAPI

        # Create database object
        database = Database(
            driver='postgresql',
            user='msdss',
            password='msdss123',
            host='localhost',
            port='5432',
            database='msdss'
        )

        # Create a users api
        # CHANGE SECRETS TO STRONG PHRASES
        users_api = UsersAPI(
            'cookie-secret',
            'jwt-secret',
            'reset-secret',
            'verification-secret',
            database=database
        )

        # Create a data api with users
        data_api = DataAPI(users_api, database=database)

        # Create a models api with users and data management
        app = ModelsAPI(
            users_api,
            models=[Model],
            database=database,
            broker_url='redis://localhost:6379/0',
            backend_url='redis://localhost:6379/0'
        )

        # Add users and data routes
        app.add_apps(users_api, data_api)

        # Run the app with app.start()
        # Try API at http://localhost:8000/docs
        # app.start()
    """
    def __init__(
        self,
        users_api=None,
        database=None,
        models=[],
        worker=None,
        broker_url=DEFAULT_BROKER_URL,
        backend_url=DEFAULT_BACKEND_URL,
        folder=DEFAULT_MODELS_FOLDER,
        models_router_settings={},
        load_env=True,
        env=ModelsDotEnv(),
        api=FastAPI(
            title='MSDSS Models API',
            version='0.0.1'
        ),
        *args, **kwargs):
        super().__init__(api=api, *args, **kwargs)

        # (ModelsAPI_env) Set env vars
        if env.exists() and load_env:
            env.load()
            broker_url = env.get('broker_url', broker_url)
            backend_url = env.get('backend_url', backend_url)
            folder = env.get('folder', folder)

        # (ModelsAPI_folder) Create folder if not exists
        if not os.path.isdir(folder):
            os.makedirs(folder)

        # (ModelsAPI_bg) Create background manager for models
        database = database if database else Database()
        data_manager = DataManager(database=database)
        models_manager = ModelsDBManager(models=models, data_manager=data_manager, folder=folder)
        worker = worker if worker else Celery(broker=broker_url, backend=backend_url)
        metadata_manager = ModelsMetadataManager(data_manager)
        bg_manager = ModelsDBBackgroundManager(worker=worker, models_manager=models_manager, metadata_manager=metadata_manager)
        models_router_settings['bg_manager'] = bg_manager
        
        # (ModelsAPI_users) Add users app if specified
        if users_api:
            models_router_settings['users_api'] = users_api

        # (ModelsAPI_router_models) Add models router
        models_router = get_models_router(**models_router_settings)
        self.add_router(models_router)

        # (ModelsAPI_router_attr) Set attributes
        self.models_api_worker = worker
        self.misc = {}
        self.misc['bg_manager'] = bg_manager

    def get_worker(self):
        """
        Get the background worker to process background tasks.

        Author
        ------
        Richard Wen <rrwen.dev@gmail.com>
        
        Example
        -------
        .. code::

            from msdss_models_api.models import Model
            from msdss_models_api import ModelsAPI

            # Create models api without users
            app = ModelsAPI(
                models=[Model],
                broker_url='redis://localhost:6379/0',
                backend_url='redis://localhost:6379/0'
            )

            # Run the background worker in a separate terminal
            worker = app.get_worker()
        """
        out = self.models_api_worker
        return out



from copy import deepcopy
from fastapi import APIRouter, Body, Depends, Query
from typing import Any, Dict, List, Literal, Optional

from .defaults import *
from .managers import *
from .models import *

async def _no_current_user():
    return None

def get_models_router(
    bg_manager,
    users_api=None,
    route_settings=DEFAULT_MODELS_ROUTE_SETTINGS,
    prefix='/models',
    tags=['models'],
    *args, **kwargs
):

    # (get_models_router_defaults) Merge defaults and user params 
    get_user = {}
    settings = deepcopy(DEFAULT_MODELS_ROUTE_SETTINGS)
    for k in settings:
        if k in route_settings:
            settings[k].update(route_settings[k])

    # (get_models_router_apply) Apply settings to obtain dependencies
    get_user = {}
    enable = {}
    for k, v in settings.items():
        get_user[k] = users_api.get_current_user(**v['_get_user']) if users_api and '_get_user' in v else _no_current_user
        del v['_get_user']
        enable[k] = v.pop('_enable')

    # (get_models_router_bg) Create bg manager func
    def get_bg_manager():
        yield bg_manager

    # (get_models_router_api) Create api router for model routes
    out = APIRouter(prefix=prefix, tags=tags, *args, **kwargs)

    # (get_models_router_cancel) Add cancel route to models router
    if enable['cancel']:
        @out.post(**settings['cancel'])
        async def cancel_model_instance_processing(
            name: str = Query(..., description='Name of the model instance'),
            bg_manager = Depends(get_bg_manager),
            user = Depends(get_user['cancel'])
        ):
            bg_manager.cancel(name)
            response = bg_manager.get_status(name)
            return response

    # (get_models_router_create) Add create route to models router
    if enable['create']:
        @out.post(**settings['create'])
        async def create_model_instance(
            name: str = Query(..., description='Name of the model instance'),
            model: str = Query(..., description='Type of model to create an instance of'),
            body: ModelCreate = Body(
                ...,
                example={
                    'title': 'Title for Model',
                    'description': 'Description for model...',
                    'tags': 'tag1 tag2 tag3',
                    'source': 'Source for model'
                }),
            bg_manager = Depends(get_bg_manager),
            user = Depends(get_user['create'])
        ):
            metadata = body.dict()
            if user:
                metadata['created_by'] = user.email
            bg_manager.create(name, model, metadata=metadata)

    # (get_models_router_delete) Add delete route to models router
    if enable['delete']:
        @out.delete(**settings['delete'])
        async def delete_model_instance(
            name: str = Query(..., description='Name of the model instance'),
            bg_manager = Depends(get_bg_manager),
            user = Depends(get_user['delete'])
        ):
            bg_manager.delete(name)

    # (get_models_router_status) Add status route to models router
    if enable['status']:
        @out.get(**settings['status'])
        async def model_instance_status(
            name: str = Query(..., description='Name of the model instance'),
            bg_manager = Depends(get_bg_manager),
            user = Depends(get_user['status'])
        ):
            response = bg_manager.get_status(name)
            return response

    # (get_models_router_input) Add input route to models router
    if enable['input']:
        @out.post(**settings['input'])
        async def model_instance_input(
            name: str = Query(..., description='Name of the model instance - the request body is used to upload JSON data under the "data" key in the form of "[{col: val, col2: val2, ...}, {col: val, col2: val2, ...}]", where each key represents a column and its corresponding value. Objects in this list should have the same keys. The "settings" key is used for specific model input settings, where each key is a setting name.'),
            data: List[Dict[str, Any]] = Body(...),
            parameters: Dict[str, Any] = {},
            bg_manager = Depends(get_bg_manager),
            user = Depends(get_user['input'])
        ):
            bg_manager.input(name, data, parameters)
            response = bg_manager.get_status(name)
            return response

    # (get_models_router_input_db) Add input db route to models router
    if enable['input_db']:
        @out.post(**settings['input_db'])
        async def model_instance_input_with_dataset(
            name: str = Query(..., description='Name of the model instance - the "settings" key in the response body is used for specific model input settings, where each key is a setting name.'),
            dataset: str = Query(..., description='Dataset name to use for initializing the model instance'),
            parameters: Dict[str, Any] = {},
            select: Optional[List[str]] = Query('*', description='Columns to include - "*" means all columns and "None" means to omit selection (useful for aggregate queries)'),
            where: Optional[List[str]] = Query(None, description='Where statements to filter data in the form of "column operator value" (e.g. "var < 3") - valid operators are: =, !=, >, >=, >, <, <=, !=, LIKE, ILIKE, NOTLIKE, NOTILIKE, CONTAINS, STARTSWITH, ENDSWITH'),
            group_by: Optional[List[str]] = Query(None, alias='group-by', description='column names to group by - should be used with aggregate and aggregate_func parameters'),
            aggregate: Optional[List[str]] = Query(None, description='Column names to aggregate with the same order as the aggregate_func parameter'),
            aggregate_func: Optional[List[str]] = Query(None, alias='aggregate-func', description='Aggregate functions in the same order as the aggregate parameter'),
            order_by: Optional[List[str]] = Query(None, alias='order-by', description='Column names to order by in the same order as parameter order_by_sort'),
            order_by_sort: Optional[List[Literal['asc', 'desc']]] = Query(None, alias='order-by-sort', description='Either "asc" for ascending or "desc" for descending order in the same order as parameter order_by'),
            limit: Optional[int] = Query(None, description='Number of items to return'),
            where_boolean: Literal['AND', 'OR'] = Query('AND', alias='where-boolean', description='Either "AND" or "OR" to combine where statements'),
            bg_manager = Depends(get_bg_manager),
            user = Depends(get_user['input_db'])
        ):
            bg_manager.input_db(
                name,
                dataset,
                settings=settings,
                select=select,
                where=where,
                group_by=group_by,
                aggregate=aggregate,
                aggregate_func=aggregate_func,
                order_by=order_by,
                order_by_sort=order_by_sort,
                limit=limit,
                where_boolean=where_boolean)
            response = bg_manager.get_status(name)
            return response

    # (get_data_router_metadata) Add metadata route to data router
    if enable['metadata']:
        @out.get(**settings['metadata'])
        async def get_model_instance_metadata(
            name: str = Query(..., description='Name of the model to get metadata for'),
            bg_manager = Depends(get_bg_manager),
            user = Depends(get_user['metadata'])
        ):
            response = bg_manager.metadata_manager.get(name=name)
            return response

    # (get_data_router_metadata) Add metadata route to data router
    if enable['metadata_update']:
        @out.put(**settings['metadata_update'])
        async def update_model_instance_metadata(
            name: str = Query(..., description='Name of the model to update metadata for. Upload user and creation/update times can not be updated.'),
            body: ModelMetadataUpdate = Body(
                ...,
                example={
                    'title': 'New Title to Replace Existing',
                    'description': 'New description to replace existing...',
                    'source': 'New source to replace existing...',
                    'tags': 'newtag1 newtag2 newtag3'
                }
            ),
            bg_manager = Depends(get_bg_manager),
            user = Depends(get_user['metadata_update'])
        ):
            bg_manager.metadata_manager.update(name=name, data=body.dict())

    # (get_models_router_output) Add output route to models router
    if enable['output']:
        @out.post(**settings['output'])
        async def model_instance_output(
            name: str = Query(..., description='Name of the model instance - the request body is used to upload JSON data under the "data" key in the form of "[{col: val, col2: val2, ...}, {col: val, col2: val2, ...}]", where each key represents a column and its corresponding value. Objects in this list should have the same keys. The "settings" key is used for specific model output settings, where each key is a setting name.'),
            data: List[Dict[str, Any]] = Body(...),
            parameters: Dict[str, Any] = {},
            bg_manager = Depends(get_bg_manager),
            user = Depends(get_user['output'])
        ):
            response = bg_manager.output(name, data, parameters).to_dict('records')
            return response

    # (get_data_router_search) Add search route to data router
    if enable['search']:
        @out.get(**settings['search'])
        async def search_model_instances(
            select: Optional[List[str]] = Query('*', description='columns to include in search - "*" means all columns and "None" means to omit selection (useful for aggregate queries).'),
            where: Optional[List[str]] = Query(None, description='Where statements to filter data in the form of "column operator value" (e.g. "dataset = test_data") - valid operators are: =, !=, >, >=, >, <, <=, !=, LIKE, ILIKE, NOTLIKE, NOTILIKE, CONTAINS, STARTSWITH, ENDSWITH'),
            order_by: Optional[List[str]] = Query(None, alias='order-by', description='column names to order by in the same order as parameter order_by_sort'),
            order_by_sort: Optional[List[Literal['asc', 'desc']]] = Query(None, alias='order-by-sort', description='Either "asc" for ascending or "desc" for descending order in the same order as parameter order_by'),
            limit: Optional[int] = Query(None, description='Number of items to return'),
            offset: Optional[int] = Query(None, description='Number of items to skip'),
            where_boolean: Literal['AND', 'OR'] = Query('AND', alias='where-boolean', description='Either "AND" or "OR" to combine where statements'),
            bg_manager = Depends(get_bg_manager),
            user = Depends(get_user['search'])
        ):
            select = None if select[0] == 'None' else select
            response = bg_manager.metadata_manager.search(
                select=select,
                where=where,
                order_by=order_by,
                order_by_sort=order_by_sort,
                limit=limit,
                offset=offset,
                where_boolean=where_boolean
            )
            return response

    # (get_models_router_update) Add update route to models router
    if enable['update']:
        @out.put(**settings['update'])
        async def update_model_instance(
            name: str = Query(..., description='Name of the model instance - the request body is used to upload JSON data under the "data" key in the form of "[{col: val, col2: val2, ...}, {col: val, col2: val2, ...}]", where each key represents a column and its corresponding value. Objects in this list should have the same keys. The "settings" key is used for specific model update settings, where each key is a setting name.'),
            data: List[Dict[str, Any]] = Body(...),
            parameters: Dict[str, Any] = {},
            bg_manager = Depends(get_bg_manager),
            user = Depends(get_user['update'])
        ):
            bg_manager.update(name, data, parameters)
            response = bg_manager.get_status(name)
            return response

    # (get_models_router_update_db) Add update db route to models router
    if enable['update_db']:
        @out.put(**settings['update_db'])
        async def update_model_instance_with_dataset(
            name: str = Query(..., description='Name of the model instance - the "settings" key in the response body is used for specific model update settings, where each key is a setting name.'),
            dataset: str = Query(..., description='Dataset name to use for updating the model instance'),
            select: Optional[List[str]] = Query('*', description='Columns to include - "*" means all columns and "None" means to omit selection (useful for aggregate queries)'),
            where: Optional[List[str]] = Query(None, description='Where statements to filter data in the form of "column operator value" (e.g. "var < 3") - valid operators are: =, !=, >, >=, >, <, <=, !=, LIKE, ILIKE, NOTLIKE, NOTILIKE, CONTAINS, STARTSWITH, ENDSWITH'),
            group_by: Optional[List[str]] = Query(None, alias='group-by', description='Column names to group by - should be used with aggregate and aggregate_func parameters'),
            aggregate: Optional[List[str]] = Query(None, description='Column names to aggregate with the same order as the aggregate_func parameter'),
            aggregate_func: Optional[List[str]] = Query(None, alias='aggregate-func', description='Aggregate functions in the same order as the aggregate parameter'),
            order_by: Optional[List[str]] = Query(None, alias='order-by', description='cClumn names to order by in the same order as parameter order_by_sort'),
            order_by_sort: Optional[List[Literal['asc', 'desc']]] = Query(None, alias='order-by-sort', description='Either "asc" for ascending or "desc" for descending order in the same order as parameter order_by'),
            limit: Optional[int] = Query(None, description='Number of items to return'),
            offset: Optional[int] = Query(None, description='Number of items to skip'),
            where_boolean: Literal['AND', 'OR'] = Query('AND', alias='where-boolean', description='Either "AND" or "OR" to combine where statements'),
            bg_manager = Depends(get_bg_manager),
            user = Depends(get_user['update_db'])
        ):
            bg_manager.update_db(
                name,
                dataset,
                settings=settings,
                select=select,
                where=where,
                group_by=group_by,
                aggregate=aggregate,
                aggregate_func=aggregate_func,
                order_by=order_by,
                order_by_sort=order_by_sort,
                limit=limit,
                offset=offset,
                where_boolean=where_boolean)
            response = bg_manager.get_status(name)
            return response

    return out


from msdss_base_dotenv import DotEnv

from .defaults import DEFAULT_DOTENV_KWARGS

class ModelsDotEnv(DotEnv):
    """
    Class to manage model environment variables.

    * Extends :class:`msdss_base_dotenv:msdss_base_dotenv.core.DotEnv`

    Parameters
    ----------
    broker_url : str
        Link to connect to a `RabbitMQ <https://www.rabbitmq.com/>`_ broker.
    backend_url : str
        Link to connect to a `RabbitMQ <https://www.rabbitmq.com/>`_ backend.
    folder : str
        Path to the folder to store saved models and instances.

    Author
    ------
    Richard Wen <rrwen.dev@gmail.com>

    Example
    -------
    .. jupyter-execute::

        from msdss_models_api.env import ModelsDotEnv
        
        # Get default env vars
        env = ModelsDotEnv()
        env.save() # save to .env file
        env.load() # load the same file
        
        # Print defaults
        print('default_env:\\n')
        for k, name in env.mappings.items():
            value = str(env.get(k))
            print(f'{name}: {value}')

        # Remove saved .env file
        env.clear()

        # Create models env with diff var names
        alt_env = ModelsDotEnv(
            broker_url='MSDSS_MODELS_BROKER_URL_B',
            backend_url='MSDSS_MODELS_BACKEND_URL_B',
            folder='MSDSS_MODELS_FOLDER_B'
        )

        # Set secret
        alt_env.set('broker_url', 'redis://localhost:6379/0')

        # Set folder
        alt_env.set('folder', './path/to/folder')
        alt_env.delete('folder')

        # Check if backend url is set
        backend_is_set = alt_env.is_set('backend_url')

        # Print custom env
        # See new env
        print('\\nalt_env:\\n')
        for k, name in alt_env.mappings.items():
            value = str(alt_env.get(k))
            print(f'{name}: {value}')
        print('backend_is_set: ' + str(backend_is_set))

        # Clear alt env files
        alt_env.clear()
    """
    def __init__(
        self,
        broker_url=DEFAULT_DOTENV_KWARGS['broker_url'],
        backend_url=DEFAULT_DOTENV_KWARGS['backend_url'],
        folder=DEFAULT_DOTENV_KWARGS['folder'],
        env_file=DEFAULT_DOTENV_KWARGS['env_file'],
        key_path=DEFAULT_DOTENV_KWARGS['key_path']):
        kwargs = locals()
        del kwargs['self']
        del kwargs['__class__']
        super().__init__(**kwargs)


DEFAULT_DOTENV_KWARGS = dict(
    env_file='./.env',
    key_path=None,
    broker_url='MSDSS_MODELS_BROKER_URL',
    backend_url='MSDSS_MODELS_BACKEND_URL',
    folder='MSDSS_MODELS_FOLDER'
)

DEFAULT_MODELS_FOLDER = './models'
DEFAULT_BROKER_URL = 'redis://localhost:6379/0'
DEFAULT_BACKEND_URL = 'redis://localhost:6379/0'

DEFAULT_METADATA_TABLE = 'model'
DEFAULT_NAME_COLUMN = 'name'
DEFAULT_UPDATE_COLUMN = 'updated_at'
DEFAULT_METADATA_COLUMNS = [
    dict(name='id', type_='Integer', primary_key=True),
    dict(name='name', type_='String', unique=True),
    ('title', 'String'),
    ('description', 'String'),
    ('tags', 'String'),
    ('source', 'String'),
    ('model', 'String'),
    ('can_input', 'Boolean'),
    ('can_output', 'Boolean'),
    ('can_update', 'Boolean'),
    ('created_by', 'String'),
    ('created_at', 'DateTime'),
    ('updated_at', 'DateTime')
]

DEFAULT_MODELS_ROUTE_SETTINGS = dict(
    cancel=dict(
        path='/{name}/cancel',
        _enable=True,
        _get_user={'superuser': True}
    ),
    create=dict(
        path='/',
        _enable=True,
        _get_user={'superuser': True}
    ),
    delete=dict(
        path='/{name}',
        _enable=True,
        _get_user={'superuser': True}
    ),
    input=dict(
        path='/{name}',
        _enable=True,
        _get_user={'superuser': True}
    ),
    input_db=dict(
        path='/{name}/data',
        _enable=True,
        _get_user={'superuser': True}
    ),
    metadata=dict(
        path='/{name}/metadata',
        tags=['metadata'],
        _enable=True,
        _get_user={}
    ),
    metadata_update=dict(
        path='/{name}/metadata',
        tags=['metadata'],
        _enable=True,
        _get_user={'superuser': True}
    ),
    output=dict(
        path='/{name}/output',
        _enable=True,
        _get_user={}
    ),
    status=dict(
        path='/{name}/status',
        _enable=True,
        _get_user={}
    ),
    search=dict(
        path='/',
        _enable=True,
        _get_user={}
    ),
    update=dict(
        path='/{name}',
        _enable=True,
        _get_user={'superuser': True}
    ),
    update_db=dict(
        path='/{name}/data',
        _enable=True,
        _get_user={'superuser': True}
    )
)


from datetime import datetime
import os
import pickle
import shutil

from msdss_base_database import Database
from msdss_data_api.managers import DataManager, MetadataManager
from shlex import split

from .defaults import *
from .handlers import *

class ModelsManager:
    """
    Class to manage :class:`msdss_models_api.models.Model` objects.
    
    Parameters
    ----------
    models : list(:class:`msdss_models_api.models.Model`) or dict(:class:`msdss_models_api.models.Model`)
        List or dict of available ``Model`` objects to use for creating and managing model instances.
        If ``list``, ensure that the class names are unique, otherwise the last object in the list takes priority.
    folder : str
        The folder path to store models in.
    handler : :class:`msdss_models_api.handlers.ModelsHandler` or None
        Handler object to manage model operations. If ``None``, then inputs for model operations will not be handled.
    suffix : str
        Suffix for pickled model object files. These files contain the model objects without inputs or loading.

    Attributes
    ----------
    models : dict(:class:`msdss_models_api.models.Model`)
        Dictionary of available models from parameter ``models``, where:

        * Each key is the class name
        * Each value is the class itself

    instances : dict(:class:`msdss_models_api.models.Model`)
        Dictionary of loaded model instances created from method ``create``.
    folder : str
        Same as parameter ``folder``.
    handler : :class:`msdss_models_api.handlers.ModelsHandler`
        Same as parameter ``handler``.
    suffix : str
        Same as parameter ``suffix``.
    
    Author
    ------
    Richard Wen <rrwen.dev@gmail.com>
    
    Example
    -------
    .. jupyter-execute::

        import tempfile
        from msdss_models_api.models import Model
        from msdss_models_api.managers import ModelsManager

        with tempfile.TemporaryDirectory() as folder_path:

            # Setup available models
            models = [Model]
            
            # Create manager
            models_manager = ModelsManager(models, folder=folder_path)

            # Create model instance
            models_manager.create('temp_model', 'Model')

            # Initialize a model instance with inputs
            train_data = [
                {'col_a': 1, 'col_b': 'a'},
                {'col_a': 2, 'col_b': 'b'}
            ]
            models_manager.input('temp_model', train_data)
            
            # Update model instance with new data
            new_data = [
                {'col_a': 2, 'col_b': 'c'},
                {'col_a': 3, 'col_b': 'd'}
            ]
            models_manager.update('temp_model', new_data)

            # Produce output from a model instance
            test_data = [
                {'col_a': 2, 'col_b': 'c'},
                {'col_a': 3, 'col_b': 'd'}
            ]
            results = models_manager.output('temp_model', test_data)

            # Delete model instance
            models_manager.delete('temp_model')
    """
    def __init__(
        self,
        models=[],
        folder=DEFAULT_MODELS_FOLDER,
        handler=ModelsHandler(),
        suffix='_base.pickle'
    ):

        # (ModelsManager_attr) Set attributes
        self.folder = folder
        self.models = {(m.__name__):m for m in models} if isinstance(models, list) else models
        self.instances = {}
        self.handler = handler
        self.suffix = suffix

        # (ModelsManager_load) Load base model instances
        self._load_base_files()

    def _get_base_file(self, name):
        """
        Get the path of the base file for the model instance.

        Parameters
        ----------
        name : str
            Unique name of the model instance. The instance is stored in ``.instances[name]``.
        
        Returns
        -------
        str
            File path for the base file of the model instance.

        Author
        ------
        Richard Wen <rrwen.dev@gmail.com>
        
        Example
        -------
        .. jupyter-execute::

            import tempfile
            from msdss_models_api.models import Model
            from msdss_models_api.managers import ModelsManager

            with tempfile.TemporaryDirectory() as folder_path:

                # Setup available models
                models = [Model]
                
                # Create manager
                models_manager = ModelsManager(models, folder=folder_path)

                # Create model instance
                models_manager.create('temp_model', 'Model')

                # Delete model instance
                path = models_manager._get_base_file('temp_model')
        """
        folder = self._get_folder(name)
        out = os.path.join(folder, name + self.suffix)
        return out

    def _get_folder(self, name):
        """
        Get the path of the subfolder for the model instance.

        Parameters
        ----------
        name : str
            Unique name of the model instance. The instance is stored in ``.instances[name]``.
        
        Returns
        -------
        str
            File path for the subfolder of the model instance.

        Author
        ------
        Richard Wen <rrwen.dev@gmail.com>
        
        Example
        -------
        .. jupyter-execute::

            import tempfile
            from msdss_models_api.models import Model
            from msdss_models_api.managers import ModelsManager

            with tempfile.TemporaryDirectory() as folder_path:

                # Setup available models
                models = [Model]
                
                # Create manager
                models_manager = ModelsManager(models, folder=folder_path)

                # Create model instance
                models_manager.create('temp_model', 'Model')

                # Delete model instance
                folder_path = models_manager._get_folder('temp_model')
        """
        out = os.path.join(self.folder, name)
        return out

    def _get_instance(self, name):
        """
        Get a model instance by name.

        Parameters
        ----------
        name : str
            Unique name of the model instance. The instance is stored in ``.instances[name]``.
        
        Author
        ------
        Richard Wen <rrwen.dev@gmail.com>
        
        Example
        -------
        .. jupyter-execute::

            import tempfile
            from msdss_models_api.models import Model
            from msdss_models_api.managers import ModelsManager

            with tempfile.TemporaryDirectory() as folder_path:

                # Setup available models
                models = [Model]
                
                # Create manager
                models_manager = ModelsManager(models, folder=folder_path)

                # Create model instance
                models_manager.create('temp_model', 'Model')

                # Delete model instance
                models_manager._get_instance('temp_model')
        """
        out = self.instances[name]
        return out

    def _get_model_name(self, name):
        """
        Get the model name of a model instance.

        Parameters
        ----------
        name : str
            Unique name of the model instance. The instance is stored in ``.instances[name]``.
        
        Author
        ------
        Richard Wen <rrwen.dev@gmail.com>
        
        Example
        -------
        .. jupyter-execute::

            import tempfile
            from msdss_models_api.models import Model
            from msdss_models_api.managers import ModelsManager

            with tempfile.TemporaryDirectory() as folder_path:

                # Setup available models
                models = [Model]
                
                # Create manager
                models_manager = ModelsManager(models, folder=folder_path)

                # Create model instance
                models_manager.create('temp_model', 'Model')

                # Delete model instance
                models_manager._get_model_name('temp_model')
        """
        instance = self._get_instance(name)
        out = type(instance).__name__
        return out

    def _get_save_file(self, name):
        """
        Get the path of the save file for the model instance without the extension.

        Parameters
        ----------
        name : str
            Unique name of the model instance. The instance is stored in ``.instances[name]``.
        
        Returns
        -------
        str
            File path for the save file of the model instance. Does not include extension.

        Author
        ------
        Richard Wen <rrwen.dev@gmail.com>
        
        Example
        -------
        .. jupyter-execute::

            import tempfile
            from msdss_models_api.models import Model
            from msdss_models_api.managers import ModelsManager

            with tempfile.TemporaryDirectory() as folder_path:

                # Setup available models
                models = [Model]
                
                # Create manager
                models_manager = ModelsManager(models, folder=folder_path)

                # Create model instance
                models_manager.create('temp_model', 'Model')

                # Delete model instance
                save_path = models_manager._get_save_file('temp_model')
        """
        folder = self._get_folder(name)
        out = os.path.join(folder, name)
        return out

    def _load_base_files(self, force=False):
        """
        Load all base models.

        Sets attribute ``.instances`` to be initialized :class:`msdss_models_api.models.Model` instances based on the subfolders in attribute ``folders``

        Parameters
        ----------
        force : bool
            Whether to force loading whether instance is in ``.instances`` attribute or not.

        Author
        ------
        Richard Wen <rrwen.dev@gmail.com>
        
        Example
        -------
        .. jupyter-execute::

            import tempfile
            from msdss_models_api.models import Model
            from msdss_models_api.managers import ModelsManager

            with tempfile.TemporaryDirectory() as folder_path:

                # Setup available models
                models = [Model]
                
                # Create manager
                models_manager = ModelsManager(models, folder=folder_path)

                # Create model instance
                models_manager.create('temp_model', 'Model')

                # Load all base models
                models_manager._load_base_files()
        """
        names = [name for name in os.listdir(self.folder) if os.path.isdir(os.path.join(self.folder, name))]
        for name in names:
            if name not in self.instances or force:
                path = self._get_base_file(name)
                with open(path, 'rb') as file:
                    self.instances[name] = pickle.load(file)

    def create(self, name, model, parameters={}):
        """
        Creates a model instance.

        * Stores the instance in attribute ``instances``
        * Creates a subfolder and base model file for the instance

        Parameters
        ----------
        name : str
            Unique name of the model instance. The instance is stored in ``.instances[name]``.
        model : str
            Name of the model to use for the instance. This is the same as the class name from attribute
        parameters : dict
            parameters to be passed to :class:`msdss_models_api.models.Model` to instantiate the instance.
        
        Author
        ------
        Richard Wen <rrwen.dev@gmail.com>
        
        Example
        -------
        .. jupyter-execute::

            import tempfile
            from msdss_models_api.models import Model
            from msdss_models_api.managers import ModelsManager

            with tempfile.TemporaryDirectory() as folder_path:

                # Setup available models
                models = [Model]
                
                # Create manager
                models_manager = ModelsManager(models, folder=folder_path)

                # Create model instance
                models_manager.create('temp_model', 'Model')
        """

        # (ModelsMAnager_create_handle) Handle create operation
        if self.handler:
            self.handler.handle_create(name, model, self.instances, self.models)
        
        # (ModelsManager_create_subfolder) Create subfolder if not exists for model
        folder = self._get_folder(name)
        if not os.path.isdir(folder):
            os.makedirs(folder)
        
        # (ModelsManager_create_obj) Create base model object
        save_file = self._get_save_file(name)
        base_model = self.models[model](file_path=save_file, **parameters)

        # (ModelsManager_create_save) Save base model object
        with open(self._get_base_file(name), 'wb') as base_file:
            pickle.dump(base_model, base_file)
            self.instances[name] = base_model

    def delete(self, name, parameters={}):
        """
        Creates a model instance.

        * Removes the instance in attribute ``instances``
        * Removes the subfolder and base model file for the instance

        Parameters
        ----------
        name : str
            Unique name of the model instance. The instance is stored in ``.instances[name]``.
        parameters : dict
            parameters to be passed to :meth:`msdss_models_api.models.Model.delete`.
        
        Author
        ------
        Richard Wen <rrwen.dev@gmail.com>
        
        Example
        -------
        .. jupyter-execute::

            import tempfile
            from msdss_models_api.models import Model
            from msdss_models_api.managers import ModelsManager

            with tempfile.TemporaryDirectory() as folder_path:

                # Setup available models
                models = [Model]
                
                # Create manager
                models_manager = ModelsManager(models, folder=folder_path)

                # Create model instance
                models_manager.create('temp_model', 'Model')

                # Initialize a model instance with inputs
                train_data = [
                    {'col_a': 1, 'col_b': 'a'},
                    {'col_a': 2, 'col_b': 'b'}
                ]
                models_manager.input('temp_model', train_data)

                # Delete model instance
                models_manager.delete('temp_model')
        """

        # (ModelsManager_delete_handle) Handle read operation
        if self.handler:
            self.handler.handle_read(name, self.instances)

        # (ModelsManager_delete_instance) Run instance deletion
        instance = self._get_instance(name)
        instance.delete(**parameters)

        # (ModelsManager_delete_folder) Delete folder
        folder = self._get_folder(name)
        shutil.rmtree(folder)
        del self.instances[name]

    def input(self, name, data, parameters={}):
        """
        Train an initial model instance and save it by adding input data.

        * Modifies ``.instance[name]`` by calling the ``input`` and ``save`` methods
        * Modifies ``.states[name]`` with action and results for inputting data into the model instance

        Parameters
        ----------
        name : str
            Unique name of the model instance. The instance is stored in ``.instances[name]``.
        data : dict or list or :class:`pandas:pandas.DataFrame`
            Data to use for initializing the model instance. Should accept a ``list`` or ``dict`` to be input in a :class:`pandas:pandas.DataFrame` or the dataframe itself.
        parameters : dict
            Keyword arguments passed to the :meth:`msdss_models_api.models.Model.input`.

        Author
        ------
        Richard Wen <rrwen.dev@gmail.com>
        
        Example
        -------
        .. jupyter-execute::

            import tempfile
            from msdss_models_api.models import Model
            from msdss_models_api.managers import ModelsManager

            with tempfile.TemporaryDirectory() as folder_path:

                # Setup available models
                models = [Model]
                
                # Create manager
                models_manager = ModelsManager(models, folder=folder_path)

                # Create model instance
                models_manager.create('temp_model', 'Model')

                # Initialize a model instance with inputs
                train_data = [
                    {'col_a': 1, 'col_b': 'a'},
                    {'col_a': 2, 'col_b': 'b'}
                ]
                models_manager.input('temp_model', train_data)
        """

        # (ModelsManager_input_handle) Handle input operation
        if self.handler:
            self.handler.handle_input(name, self.instances)

        # (ModelsManager_input_run) Initialize model with input data
        instance = self._get_instance(name)
        instance.input(data, **parameters)
        instance.save()

    def output(self, name, data, parameters={}):
        """
        Get the output of a model instance.

        Parameters
        ----------
        name : str
            Unique name of the model instance. The instance is stored in ``.instances[name]``.
        data : dict or list or :class:`pandas:pandas.DataFrame`
            Data to use for obtaining the model instance output. Should accept a ``list`` or ``dict`` to be input in a :class:`pandas:pandas.DataFrame` or the dataframe itself.
        parameters : dict
            Keyword arguments passed to the :meth:`msdss_models_api.models.Model.output`.

        Returns
        -------
        :class:`pandas:pandas.DataFrame`
            Output data from the ``name`` model instance using the input data from parameter ``data``.

        Author
        ------
        Richard Wen <rrwen.dev@gmail.com>
        
        Example
        -------
        .. jupyter-execute::

            import tempfile
            from msdss_models_api.models import Model
            from msdss_models_api.managers import ModelsManager

            with tempfile.TemporaryDirectory() as folder_path:

                # Setup available models
                models = [Model]
                
                # Create manager
                models_manager = ModelsManager(models, folder=folder_path)

                # Create model instance
                models_manager.create('temp_model', 'Model')

                # Initialize a model instance with inputs
                train_data = [
                    {'col_a': 1, 'col_b': 'a'},
                    {'col_a': 2, 'col_b': 'b'}
                ]
                models_manager.input('temp_model', train_data)

                # Produce output from a model instance
                test_data = [
                    {'col_a': 2, 'col_b': 'c'},
                    {'col_a': 3, 'col_b': 'd'}
                ]
                results = models_manager.output('temp_model', test_data)
        """

        # (ModelsManager_output_handle) Handle output operation
        if self.handler:
            self.handler.handle_output(name, self.instances)

        # (ModelsManager_output_run) Get model output
        instance = self._get_instance(name)
        instance.load()
        out = instance.output(data, **parameters)
        return out
    
    def update(self, name, data, parameters={}):
        """
        Update a model instance with new data.

        * Modifies ``.instance[name]`` by calling the ``update`` and ``save`` methods
        * Modifies ``.states[name]`` with action and results for updating the model instance

        Parameters
        ----------
        name : str
            Unique name of the model instance. The instance is stored in ``.instances[name]``.
        data : dict or list or :class:`pandas:pandas.DataFrame`
            Data to use for updating the model instance. Should accept a ``list`` or ``dict`` to be input in a :class:`pandas:pandas.DataFrame` or the dataframe itself.
        parameters : dict
            Keyword arguments passed to the :meth:`msdss_models_api.models.Model.update`.

        Author
        ------
        Richard Wen <rrwen.dev@gmail.com>
        
        Example
        -------
        .. jupyter-execute::

            import tempfile
            from msdss_models_api.models import Model
            from msdss_models_api.managers import ModelsManager

            with tempfile.TemporaryDirectory() as folder_path:

                # Setup available models
                models = [Model]
                
                # Create manager
                models_manager = ModelsManager(models, folder=folder_path)

                # Create model instance
                models_manager.create('temp_model', 'Model')

                # Initialize a model instance with inputs
                train_data = [
                    {'col_a': 1, 'col_b': 'a'},
                    {'col_a': 2, 'col_b': 'b'}
                ]
                models_manager.input('temp_model', train_data)

                # Update model instance with new data
                new_data = [
                    {'col_a': 2, 'col_b': 'c'},
                    {'col_a': 3, 'col_b': 'd'}
                ]
                models_manager.update('temp_model', new_data)
        """

        # (ModelsManager_update_handle) Handle update operation
        if self.handler:
            self.handler.handle_update(name, self.instances)

        # (ModelsManager_update_run) Update model instance
        instance = self._get_instance(name)
        instance.load()
        instance.update(data, **parameters)
        instance.save()

class ModelsBackgroundManager:
    """
    Class to manage :class:`msdss_models_api.models.Model` background processes using a :class:`msdss_models_api.managers.ModelsManager`.
    
    * Note: it is expected that the model outputs and creation are not long running processes

    Parameters
    ----------
    worker : :class:`celery:celery.Celery`
        A ``celery`` app object to manage background tasks.
    models_manager : :class:`msdss_models_api.managers.ModelsManager`
        A models manager object to manage models.
        The handler for the models manager will be disabled (set to ``None``) as model operations will be handled with attribute ``models_handler`` instead.
    handler : :class:`msdss_models_api.handlers.ModelsBackgroundHandler`
        Handler object to manage background operations.
    metadata_manager : :class:`msdss_models_api.managers.ModelsMetadataManager` or ``None``
        A metadata manager object. If ``None``, metadata will not be managed based on model operations.

    Attributes
    ----------
    worker : :class:`celery:celery.Celery` or None
        Same as parameter ``worker``.
    models_manager : dict(:class:`msdss_models_api.manager.ModelsManager`)
        Same as parameter ``models_manager``.
    tasks : dict
        Dictionary of background tasks from the ``worker`` object with the following keys:

        * ``input`` (func): background task to initialize models with input data
        * ``update`` (func): background task to update models

    states : dict
        Dictionary of processing states for each instance, consisting of the following keys:

        * ``task`` (str): the action that the process is performing - one of: ``INPUT``, ``UPDATE``
        * ``result`` (:class:`celery:celery.result.AsyncResult`): celery async object for getting states, ids, etc (see `celery.result <https://docs.celeryproject.org/en/stable/reference/celery.result.html#celery.result.AsyncResult>`_)
        * ``started_at`` (:class:`datetime.datetime`): datetime object for when the task was started
    
    models_handler : :class:`msdss_models_api.handlers.ModelsHandler`
        Handler extracted from parameter ``models_manager``.
    handler : :class:`msdss_models_api.handlers.ModelsBackgroundHandler`
        Same as parameter ``handler``.
    metadata_manager : :class:`msdss_models_api.managers.ModelsMetadataManager`
        Same as parameter ``metadata_manager``.
    
    Author
    ------
    Richard Wen <rrwen.dev@gmail.com>
    
    Example
    -------
    .. code::

        import tempfile
        from celery import Celery
        from msdss_models_api.models import Model
        from msdss_models_api.managers import *
        from pprint import pprint

        with tempfile.TemporaryDirectory() as folder_path:

            # Setup available models
            models = [Model]
            
            # Create manager
            models_manager = ModelsManager(models, folder=folder_path)

            # Create model instance
            models_manager.create('temp_model', 'Model')

            # Create background manager
            worker = Celery(broker='redis://localhost:6379/0', backend='redis://localhost:6379/0') # rabbitmq
            bg_manager = ModelsBackgroundManager(worker, models_manager)

            # Initialize a model instance with inputs as a background process
            train_data = [
                {'col_a': 1, 'col_b': 'a'},
                {'col_a': 2, 'col_b': 'b'}
            ]
            bg_manager.input('temp_model', train_data)

            # Check status
            status = bg_manager.get_status('temp_model')
            pprint(status)
            
            # Update model instance with new data as background process
            new_data = [
                {'col_a': 2, 'col_b': 'c'},
                {'col_a': 3, 'col_b': 'd'}
            ]
            bg_manager.update('temp_model', new_data)

            # Produce output from a model instance
            test_data = [
                {'col_a': 2, 'col_b': 'c'},
                {'col_a': 3, 'col_b': 'd'}
            ]
            results = bg_manager.output('temp_model', test_data)

            # Delete model instance
            bg_manager.delete('temp_model')
    """
    def __init__(self, worker, models_manager, handler=ModelsBackgroundHandler(), metadata_manager=None):

        # (ModelsBackgroundManager_models_manager) Setup models manager and disable handler
        self.models_manager = models_manager
        self.models_handler = self.models_manager.handler
        self.models_manager.handler = None
        self.metadata_manager = metadata_manager

        # (ModelsBackgroundManager_attr) Set other attributes
        self.worker = worker
        self.tasks = {}
        self.states = {}
        self.handler = handler

        # (ModelsBackgroundManager_task_input) Define input task
        @self.worker.task
        def input(name, data, parameters={}):
            models_manager._load_base_files()
            models_manager.input(name, data, parameters)
            if metadata_manager:
                metadata_manager.updated_at(name)
        self.tasks['input'] = input

        # (ModelsBackgroundManager_task_update) Define update task
        @self.worker.task
        def update(name, data, parameters={}):
            models_manager._load_base_files()
            models_manager.update(name, data, parameters)
            if metadata_manager:
                metadata_manager.updated_at(name)
        self.tasks['update'] = update

    def _add_task(self, task, name, *args, **kwargs):
        """
        Add a background task for a model instance.

        * Sets attribute ``.states`` with a key referring to the model instance ``name``

        Parameters
        ----------
        task : str
            The name of the task in attribute ``.tasks``.
        name : str
            The name of the model instance to add tasks for.
        *args, **kwargs
            Additional arguments passed to the associated task function in attribute ``.tasks``.

        Author
        ------
        Richard Wen <rrwen.dev@gmail.com>
        
        Example
        -------
        .. code::

            import tempfile
            from celery import Celery
            from msdss_models_api.models import Model
            from msdss_models_api.managers import *

            with tempfile.TemporaryDirectory() as folder_path:

                # Setup available models
                models = [Model]
                
                # Create manager
                models_manager = ModelsManager(models, folder=folder_path)

                # Create model instance
                models_manager.create('temp_model', 'Model')

                # Create background manager
                worker = Celery(broker='redis://localhost:6379/0', backend='redis://localhost:6379/0') # rabbitmq
                bg_manager = ModelsBackgroundManager(worker, models_manager)

                # Initialize a model instance with inputs as a background process
                train_data = [
                    {'col_a': 1, 'col_b': 'a'},
                    {'col_a': 2, 'col_b': 'b'}
                ]
                bg_manager._add_task('input', 'temp_model', train_data)
        """
        result = self.tasks[task].delay(name, *args, **kwargs)
        self.states[name] = {
            'task': task.upper(),
            'result': result,
            'started_at': datetime.now()
        }

    def cancel(self, name):
        """
        Cancel background task for model instance.

        Parameters
        ----------
        name : str
            The name of the model instance to cancel tasks for.

        Author
        ------
        Richard Wen <rrwen.dev@gmail.com>
        
        Example
        -------
        .. code::

            import tempfile
            from celery import Celery
            from msdss_models_api.models import Model
            from msdss_models_api.managers import *

            with tempfile.TemporaryDirectory() as folder_path:

                # Setup available models
                models = [Model]
                
                # Create manager
                models_manager = ModelsManager(models, folder=folder_path)

                # Create model instance
                models_manager.create('temp_model', 'Model')

                # Create background manager
                worker = Celery(broker='redis://localhost:6379/0', backend='redis://localhost:6379/0') # rabbitmq
                bg_manager = ModelsBackgroundManager(worker, models_manager)

                # Initialize a model instance with inputs as a background process
                train_data = [
                    {'col_a': 1, 'col_b': 'a'},
                    {'col_a': 2, 'col_b': 'b'}
                ]
                bg_manager.input('temp_model', train_data)
                
                # Cancel any bg tasks for model instance
                bg_manager.cancel('temp_model')
        """
        self.handler.handle_cancel(name, self.states)
        result = self.state[name]['result']
        result.revoke(terminate=True)

    def create(self, name, model, metadata={}, *args, **kwargs):
        """
        Create a model instance.

         * Runs :meth:`msdss_models_api.managers.ModelsManager.create`

        Parameters
        ----------
        name : str
            See parameter ``name`` in :meth:`msdss_models_api.managers.ModelsManager.create`.
        model : str
            See parameter ``model`` in :meth:`msdss_models_api.managers.ModelsManager.create`.
        metadata : dict
            Metadata to add for the model if attribute ``.metadata_manager`` is not ``None``.
        *args, **kwargs
            Additional arguments passed to :meth:`msdss_models_api.managers.ModelsManager.create`.

        Author
        ------
        Richard Wen <rrwen.dev@gmail.com>
        
        Example
        -------
        .. code::

            import tempfile
            from celery import Celery
            from msdss_models_api.models import Model
            from msdss_models_api.managers import *

            with tempfile.TemporaryDirectory() as folder_path:

                # Setup available models
                models = [Model]
                
                # Create manager
                models_manager = ModelsManager(models, folder=folder_path)

                # Create background manager
                worker = Celery(broker='redis://localhost:6379/0', backend='redis://localhost:6379/0') # rabbitmq
                bg_manager = ModelsBackgroundManager(worker, models_manager)

                # Create model instance
                bg_manager.create('temp_model', 'Model')
        """

        # (ModelsBackgroundManager_create_model) Create model
        self.models_manager._load_base_files()
        self.models_handler.handle_create(name, model, self.models_manager.instances, self.models_manager.models)
        self.models_manager.create(name, model, *args, **kwargs)

        # (ModelsBackgroundManager_create_metadata) Add initial metadata for model
        if self.metadata_manager:
            instance = self.models_manager._get_instance(name)
            metadata['name'] = name
            metadata['created_at'] = datetime.now()
            metadata['updated_at'] = datetime.now()
            metadata['model'] = instance.metadata['name']
            metadata['can_input'] = instance.metadata['can_input']
            metadata['can_output'] = instance.metadata['can_output']
            metadata['can_update'] = instance.metadata['can_update']
            self.metadata_manager.create(name=name, data=metadata)

    def delete(self, name, *args, **kwargs):
        """
        Delete a model instance and stop any associated background tasks.

        * Runs :meth:`msdss_models_api.managers.ModelsManager.delete` and :meth:`msdss_models_api.managers.ModelsBackgroundManager.cancel`

        Parameters
        ----------
        name : str
            See parameter ``name`` in :meth:`msdss_models_api.managers.ModelsManager.delete`.
        *args, **kwargs
            Additional arguments passed to :meth:`msdss_models_api.managers.ModelsManager.delete`.

        Author
        ------
        Richard Wen <rrwen.dev@gmail.com>
        
        Example
        -------
        .. code::

            import tempfile
            from celery import Celery
            from msdss_models_api.models import Model
            from msdss_models_api.managers import *

            with tempfile.TemporaryDirectory() as folder_path:

                # Setup available models
                models = [Model]
                
                # Create manager
                models_manager = ModelsManager(models, folder=folder_path)

                # Create model instance
                models_manager.create('temp_model', 'Model')

                # Create background manager
                worker = Celery(broker='redis://localhost:6379/0', backend='redis://localhost:6379/0') # rabbitmq
                bg_manager = ModelsBackgroundManager(worker, models_manager)

                # Initialize a model instance with inputs as a background process
                train_data = [
                    {'col_a': 1, 'col_b': 'a'},
                    {'col_a': 2, 'col_b': 'b'}
                ]
                bg_manager.input('temp_model', train_data)
                
                # Delete the model instance and any associated bg tasks
                bg_manager.delete('temp_model')
        """
        self.models_manager._load_base_files()
        self.models_handler.handle_read(name, self.models_manager.instances)
        if name in self.states:
            status = self.states[name]['result'].state
            if status in ['PENDING', 'STARTED', 'RETRY']:
                self.cancel(name)
        self.models_manager.delete(name, *args, **kwargs)
        self.metadata_manager.delete(name)

    def get_status(self, name):
        """
        Get the status of a model instance.

        Parameters
        ----------
        name : str
            See parameter ``name`` in :meth:`msdss_models_api.managers.ModelsManager.load`.

        Returns
        -------
        dict
            A dictionary representing the status of the model instance with ``name``, containing the following keys:

            * ``task`` (str): the processing task of the model instance
            * ``status`` (str): the processing status of the model instance 
            * ``started_at`` (:class:`datetime.datetime): when the process was started

            If the model instance is not processing, it will return a dict of the status only ``{'status': 'IDLE'}``.

        Author
        ------
        Richard Wen <rrwen.dev@gmail.com>
        
        Example
        -------
        .. code::

            import tempfile
            from celery import Celery
            from msdss_models_api.models import Model
            from msdss_models_api.managers import *
            from pprint import pprint

            with tempfile.TemporaryDirectory() as folder_path:

                # Setup available models
                models = [Model]
                
                # Create manager
                models_manager = ModelsManager(models, folder=folder_path)

                # Create model instance
                models_manager.create('temp_model', 'Model')

                # Create background manager
                worker = Celery(broker='redis://localhost:6379/0', backend='redis://localhost:6379/0') # rabbitmq
                bg_manager = ModelsBackgroundManager(worker, models_manager)

                # Initialize a model instance with inputs as a background process
                train_data = [
                    {'col_a': 1, 'col_b': 'a'},
                    {'col_a': 2, 'col_b': 'b'}
                ]
                bg_manager.input('temp_model', train_data)
                
                # Get the status of the model instance
                status = bg_manager.get_status('temp_model')
                pprint(status)
        """
        if name in self.states:
            state = self.states[name]
            out = {
                'task': state['task'],
                'status': state['result'].state,
                'started_at': state['started_at']
            }
        else:
            out = {
                'status': 'IDLE'
            }
        return out

    def input(self, name, *args, **kwargs):
        """
        Initialize a model instance with data as a background task.

        Runs :meth:`msdss_models_api.managers.ModelsManager.input`.

        Parameters
        ----------
        name : str
            See parameter ``name`` in :meth:`msdss_models_api.managers.ModelsManager.input`.
        *args, **kwargs
            Additional arguments passed to the :meth:`msdss_models_api.managers.ModelsManager.input`.

        Author
        ------
        Richard Wen <rrwen.dev@gmail.com>
        
        Example
        -------
        .. code::

            import tempfile
            from celery import Celery
            from msdss_models_api.models import Model
            from msdss_models_api.managers import *

            with tempfile.TemporaryDirectory() as folder_path:

                # Setup available models
                models = [Model]
                
                # Create manager
                models_manager = ModelsManager(models, folder=folder_path)

                # Create model instance
                models_manager.create('temp_model', 'Model')

                # Create background manager
                worker = Celery(broker='redis://localhost:6379/0', backend='redis://localhost:6379/0') # rabbitmq
                bg_manager = ModelsBackgroundManager(worker, models_manager)

                # Initialize a model instance with inputs as a background process
                train_data = [
                    {'col_a': 1, 'col_b': 'a'},
                    {'col_a': 2, 'col_b': 'b'}
                ]
                bg_manager.input('temp_model', train_data)
        """
        self.models_manager._load_base_files()
        self.models_handler.handle_input(name, self.models_manager.instances)
        self.handler.handle_processing(name, self.states)
        self._add_task('input', name, *args, **kwargs)

    def output(self, name, *args, **kwargs):
        """
        Get output from a model, only if a model is not processing.

        Runs :meth:`msdss_models_api.managers.ModelsManager.output`.

        Parameters
        ----------
        name : str
            See parameter ``name`` in :meth:`msdss_models_api.managers.ModelsManager.output`.
        *args, **kwargs
            Additional arguments passed to the :meth:`msdss_models_api.managers.ModelsManager.output`.

        Author
        ------
        Richard Wen <rrwen.dev@gmail.com>
        
        Example
        -------
        .. code::

            import tempfile
            from celery import Celery
            from msdss_models_api.models import Model
            from msdss_models_api.managers import *

            with tempfile.TemporaryDirectory() as folder_path:

                # Setup available models
                models = [Model]
                
                # Create manager
                models_manager = ModelsManager(models, folder=folder_path)

                # Create model instance
                models_manager.create('temp_model', 'Model')

                # Create background manager
                worker = Celery(broker='redis://localhost:6379/0', backend='redis://localhost:6379/0') # rabbitmq
                bg_manager = ModelsBackgroundManager(worker, models_manager)

                # Initialize a model instance with inputs as a background process
                train_data = [
                    {'col_a': 1, 'col_b': 'a'},
                    {'col_a': 2, 'col_b': 'b'}
                ]
                bg_manager.input('temp_model', train_data)
                
                # Get output from model instance
                test_data = [
                    {'col_a': 2, 'col_b': 'c'},
                    {'col_a': 3, 'col_b': 'd'}
                ]
                bg_manager.output('temp_model', new_data)
        """
        self.models_manager._load_base_files()
        self.models_handler.handle_output(name, self.models_manager.instances)
        self.handler.handle_processing(name, self.states)
        return self.models_manager.output(name, *args, **kwargs)

    def start(self, *args, worker_kwargs={}, **kwargs):
        """
        Start the background worker to process background tasks.

        Parameters
        ----------
        worker_kwargs : dict
            Keyword arguments for :class:`celery:celery.apps.worker`.
        *args, **kwargs
            Additional arguments passed to :meth:`celery:celery.apps.worker.start`.

        Author
        ------
        Richard Wen <rrwen.dev@gmail.com>
        
        Example
        -------
        .. code::

            import tempfile
            from celery import Celery
            from msdss_models_api.models import Model
            from msdss_models_api.managers import *

            with tempfile.TemporaryDirectory() as folder_path:

                # Setup available models
                models = [Model]
                
                # Create manager
                models_manager = ModelsManager(models, folder=folder_path)

                # Create model instance
                models_manager.create('temp_model', 'Model')

                # Create background manager
                worker = Celery(broker='redis://localhost:6379/0', backend='redis://localhost:6379/0') # rabbitmq
                bg_manager = ModelsBackgroundManager(worker, models_manager)

                # Initialize a model instance with inputs as a background process
                train_data = [
                    {'col_a': 1, 'col_b': 'a'},
                    {'col_a': 2, 'col_b': 'b'}
                ]
                bg_manager._add_task('input', 'temp_model', train_data)

                # Start worker
                bg_manager.start()
        """
        self.worker.Worker(**worker_kwargs).start(*args, **kwargs)
    
    def update(self, name, *args, **kwargs):
        """
        Update a model instance with new data as a background task.

        Runs :meth:`msdss_models_api.managers.ModelsManager.update`.

        Parameters
        ----------
        name : str
            See parameter ``name`` in :meth:`msdss_models_api.managers.ModelsManager.update`.
        *args, **kwargs
            Additional arguments passed to the :meth:`msdss_models_api.managers.ModelsManager.update`.

        Author
        ------
        Richard Wen <rrwen.dev@gmail.com>
        
        Example
        -------
        .. code::

            import tempfile
            from celery import Celery
            from msdss_models_api.models import Model
            from msdss_models_api.managers import *

            with tempfile.TemporaryDirectory() as folder_path:

                # Setup available models
                models = [Model]
                
                # Create manager
                models_manager = ModelsManager(models, folder=folder_path)

                # Create model instance
                models_manager.create('temp_model', 'Model')

                # Create background manager
                worker = Celery(broker='redis://localhost:6379/0', backend='redis://localhost:6379/0') # rabbitmq
                bg_manager = ModelsBackgroundManager(worker, models_manager)

                # Initialize a model instance with inputs as a background process
                train_data = [
                    {'col_a': 1, 'col_b': 'a'},
                    {'col_a': 2, 'col_b': 'b'}
                ]
                bg_manager.input('temp_model', train_data)
                
                # Update model instance with new data as background process
                new_data = [
                    {'col_a': 2, 'col_b': 'c'},
                    {'col_a': 3, 'col_b': 'd'}
                ]
                bg_manager.update('temp_model', new_data)
        """
        self.models_manager._load_base_files()
        self.models_handler.handle_update(name, self.models_manager.instances)
        self.handler.handle_processing(name, self.states)
        self._add_task('update', name, *args, **kwargs)

class ModelsDBManager(ModelsManager):
    """
    Class to manage :class:`msdss_models_api.models.Model` objects with added methods for processing models with data from a database.

    * Inherits from :class:`msdss_models_api.models.ModelsManager`
    
    Parameters
    ----------
    models : list(:class:`msdss_models_api.models.Model`)
        List of available ``Model`` objects to use for creating and managing model instances.
        Ensure that the class names are unique, otherwise the last object takes priority.
    data_manager : :class:`msdss_data_api.managers.DataManager`
        A data manager object for managing data in and out of a database.
    *args, **kwargs
        Additional arguments for :class:`msdss_models_api.models.ModelsManager`
    
    Author
    ------
    Richard Wen <rrwen.dev@gmail.com>
    
    Example
    -------
    .. jupyter-execute::

        import tempfile
        from msdss_base_database import Database
        from msdss_data_api.managers import DataManager
        from msdss_models_api.models import Model
        from msdss_models_api.managers import ModelsDBManager

        with tempfile.TemporaryDirectory() as folder_path:

            # Setup available models and database
            database = Database()
            models = [Model]
            
            # Create manager
            data_manager = DataManager(database=database)
            models_manager = ModelsDBManager(models, data_manager, folder=folder_path)

            # Create model instance
            models_manager.create('temp_model', 'Model')

            # Add training data to database
            train_data = [
                {'col_a': 1, 'col_b': 'a'},
                {'col_a': 2, 'col_b': 'b'}
            ]
            database.insert('models_test', train_data)

            # Initialize a model instance with inputs from database
            models_manager.input_db('temp_model', 'models_test')
            
            # Update model instance with new data
            new_data = [
                {'col_a': 3, 'col_b': 'c'},
                {'col_a': 4, 'col_b': 'd'}
            ]
            database.insert('models_test', new_data)
            models_manager.update_db('temp_model', 'models_test', where=['col_a > 2'])

            # Delete test table
            database.drop_table('models_test')
    """
    def __init__(self, models=[], data_manager=DataManager(), *args, **kwargs):
        super().__init__(models=models, *args, **kwargs)
        self.data_manager = data_manager
    
    def input_db(self, name, dataset, parameters={}, *args, **kwargs):
        """
        Initialize a model instance with data from the database.

        Parameters
        ----------
        name : str
            Unique name of the model instance. The instance is stored in ``.instances[name]``.
        dataset : str
            Name of the table from the database. See parameter ``dataset`` :meth:`msdss_data_api.managers.DataManager.get`.
        parameters : dict
            Keyword arguments passed to the :meth:`msdss_models_api.models.Model.input`.
        *args, **kwargs
            Additional arguments passed to :meth:`msdss_data_api.managers.DataManager.get`.

        Author
        ------
        Richard Wen <rrwen.dev@gmail.com>
        
        Example
        -------
        .. jupyter-execute::

            import tempfile
            from msdss_base_database import Database
            from msdss_data_api.managers import DataManager
            from msdss_models_api.models import Model
            from msdss_models_api.managers import ModelsDBManager

            with tempfile.TemporaryDirectory() as folder_path:

                # Setup available models and database
                database = Database()
                models = [Model]
                
                # Create manager
                data_manager = DataManager(database=database)
                models_manager = ModelsDBManager(models, data_manager, folder=folder_path)

                # Create model instance
                models_manager.create('temp_model', 'Model')

                # Add training data to database
                train_data = [
                    {'col_a': 1, 'col_b': 'a'},
                    {'col_a': 2, 'col_b': 'b'}
                ]
                database.insert('models_test', train_data)

                # Initialize a model instance with inputs from database
                models_manager.input_db('temp_model', 'models_test')

                # Delete test table
                database.drop_table('models_test')
        """
        data = self.data_manager.get(dataset, *args, **kwargs)
        self.input(name, data, parameters)

    def update_db(self, name, dataset, parameters={}, *args, **kwargs):
        """
        Update a model instance with data from the database.

        Parameters
        ----------
        name : str
            Unique name of the model instance. The instance is stored in ``.instances[name]``.
        dataset : str
            Name of the table from the database. See parameter ``dataset`` :meth:`msdss_data_api.managers.DataManager.get`.
        parameters : dict
            Keyword arguments passed to the :meth:`msdss_models_api.models.Model.update`.
        *args, **kwargs
            Additional arguments passed to :meth:`msdss_data_api.managers.DataManager.get`.

        Author
        ------
        Richard Wen <rrwen.dev@gmail.com>
        
        Example
        -------
        .. jupyter-execute::

            import tempfile
            from msdss_base_database import Database
            from msdss_data_api.managers import DataManager
            from msdss_models_api.models import Model
            from msdss_models_api.managers import ModelsDBManager

            with tempfile.TemporaryDirectory() as folder_path:

                # Setup available models and database
                database = Database()
                models = [Model]
                
                # Create manager
                data_manager = DataManager(database=database)
                models_manager = ModelsDBManager(models, data_manager, folder=folder_path)

                # Create model instance
                models_manager.create('temp_model', 'Model')

                # Add training data to database
                train_data = [
                    {'col_a': 1, 'col_b': 'a'},
                    {'col_a': 2, 'col_b': 'b'}
                ]
                database.insert('models_test', train_data)

                # Initialize a model instance with inputs from database
                models_manager.input_db('temp_model', 'models_test')

                # Update model instance with new data
                new_data = [
                    {'col_a': 3, 'col_b': 'c'},
                    {'col_a': 4, 'col_b': 'd'}
                ]
                database.insert('models_test', new_data)
                models_manager.update_db('temp_model', 'models_test', where=['col_a > 2'])

                # Delete test table
                database.drop_table('models_test')
        """
        data = self.data_manager.get(dataset, *args, **kwargs)
        self.update(name, data, parameters)

class ModelsDBBackgroundManager(ModelsBackgroundManager):
    """
    Class to manage :class:`msdss_models_api.models.Model` background processes using a :class:`msdss_models_api.managers.ModelsManager`.
    
    * Inherits from :class:`msdss_models_api.managers.ModelsBackgroundManager`

    Parameters
    ----------
    models_manager : :class:`msdss_models_api.managers.ModelsDBManager`
        A models DB manager object to manage models that can receive data from a database.
        The handler and data handler for the models manager will be disabled as model operations will be handled with parameter ``handler`` instead.
    *args, **kwargs
        Additional arguments passed to :class:`msdss_models_api.managers.ModelsBackgroundManager`.

    Attributes
    ----------
    models_manager : dict(:class:`msdss_models_api.manager.ModelsDBManager`)
        Same as parameter ``models_manager``.
    tasks : dict
        Dictionary of background tasks from the ``worker`` object with the following keys (in addition to the ones in attribute ``tasks`` of :class:`msdss_models_api.managers.ModelsBackgroundManager`):

        * ``input_db`` (func): background task to initialize models with input data from a database
        * ``update_db`` (func): background task to update models using data from a database
    
    Author
    ------
    Richard Wen <rrwen.dev@gmail.com>
    
    Example
    -------
    .. code::

        import tempfile
        from celery import Celery
        from msdss_base_database import Database
        from msdss_data_api.managers import DataManager
        from msdss_models_api.models import Model
        from msdss_models_api.managers import *
        from pprint import pprint

        with tempfile.TemporaryDirectory() as folder_path:

            # Setup available models
            models = [Model]
            
            # Create manager
            database = Database()
            data_manager = DataManager(database=database)
            models_manager = ModelsDBManager(models, data_manager, folder=folder_path)

            # Create model instance
            models_manager.create('temp_model', 'Model')

            # Create background manager
            worker = Celery(broker='redis://localhost:6379/0', backend='redis://localhost:6379/0') # rabbitmq
            bg_manager = ModelsDBBackgroundManager(worker, models_manager)

             # Add training data to database
            train_data = [
                {'col_a': 1, 'col_b': 'a'},
                {'col_a': 2, 'col_b': 'b'}
            ]
            database.insert('models_test', train_data)

            # Initialize a model instance with inputs from database
            bg_manager.input_db('temp_model', 'models_test')
            
            # Update model instance with new data
            new_data = [
                {'col_a': 3, 'col_b': 'c'},
                {'col_a': 4, 'col_b': 'd'}
            ]
            database.insert('models_test', new_data)
            bg_manager.update_db('temp_model', 'models_test', where=['col_a > 2'])
    """
    def __init__(self, models_manager, *args, **kwargs):

        # (ModelsDBBackgroundManager_init) Initialize inherited class
        super().__init__(models_manager=models_manager, *args, **kwargs)

        # (ModelsDBBackgroundManager_data_manager) Setup data manager
        self.data_handler = self.models_manager.data_manager.handler
        self.models_manager.data_manager.handler = None
        
        # (ModelsDBBackgroundManager_task_input) Define input db task
        @self.worker.task
        def input_db(name, dataset, parameters={}, *args, **kwargs):
            models_manager.input_db(name, dataset, parameters, *args, **kwargs)
            if self.metadata_manager:
                self.metadata_manager.updated_at(name)
        self.tasks['input_db'] = input_db

        # (ModelsDBBackgroundManager_task_update) Define update db task
        @self.worker.task
        def update_db(name, dataset, parameters={}, *args, **kwargs):
            models_manager.update_db(name, dataset, parameters, *args, **kwargs)
            if self.metadata_manager:
                self.metadata_manager.updated_at(name)
        self.tasks['update_db'] = update_db

    def input_db(self, name, dataset, parameters={}, where=None, *args, **kwargs):
        """
        Initialize a model instance with data from the database.

        Parameters
        ----------
        name : str
            Unique name of the model instance. The instance is stored in ``.instances[name]``.
        dataset : str
            Name of the table from the database. See parameter ``dataset`` :meth:`msdss_data_api.managers.DataManager.get`.
        parameters : dict
            Keyword arguments passed to the :meth:`msdss_models_api.models.Model.input`.
        where : list(str)
            See parameter ``where`` in :meth:`msdss_data_api.managers.DataManager.get`.
        *args, **kwargs
            Additional arguments passed to :meth:`msdss_data_api.managers.DataManager.get`.

        Author
        ------
        Richard Wen <rrwen.dev@gmail.com>
        
        Example
        -------
        .. code::

            import tempfile
            from celery import Celery
            from msdss_base_database import Database
            from msdss_data_api.managers import DataManager
            from msdss_models_api.models import Model
            from msdss_models_api.managers import *
            from pprint import pprint

            with tempfile.TemporaryDirectory() as folder_path:

                # Setup available models
                models = [Model]
                
                # Create manager
                database = Database()
                data_manager = DataManager(database=database)
                models_manager = ModelsDBManager(models, data_manager, folder=folder_path)

                # Create model instance
                models_manager.create('temp_model', 'Model')

                # Create background manager
                worker = Celery(broker='redis://localhost:6379/0', backend='redis://localhost:6379/0') # rabbitmq
                bg_manager = ModelsDBBackgroundManager(worker, models_manager)

                # Add training data to database
                train_data = [
                    {'col_a': 1, 'col_b': 'a'},
                    {'col_a': 2, 'col_b': 'b'}
                ]
                database.insert('models_test', train_data)

                # Initialize a model instance with inputs from database
                bg_manager.input_db('temp_model', 'models_test')
        """
        self.models_manager._load_base_files()
        
        # (ModelsDBBackgroundManager_input_db_data) Handle database read and query
        where = [split(w) for w in where] if where else where
        self.data_handler.handle_read(dataset)
        self.data_handler.handle_where(where)

        # (ModelsDBBackgroundManager_input_db_handle) Handle model input and processing
        self.models_handler.handle_input(name, self.models_manager.instances)
        self.handler.handle_processing(name, self.states)

        # (ModelsDBBackgroundManager_input_db_add) Add input db task
        self._add_task('input_db', name, dataset, parameters=parameters, where=where, *args, **kwargs)

    def update_db(self, name, dataset, parameters={}, where=None, *args, **kwargs):
        """
        Update a model instance with new data from the database as a background task.

        Runs :meth:`msdss_models_api.managers.ModelsDBManager.update_db`.

        Parameters
        ----------
        name : str
            See parameter ``name`` in :meth:`msdss_models_api.managers.ModelsManager.update`.
        dataset : str
            Name of the table from the database. See parameter ``dataset`` :meth:`msdss_data_api.managers.DataManager.get`.
        parameters : dict
            Keyword arguments passed to the :meth:`msdss_models_api.models.Model.update`.
        where : list(str)
            See parameter ``where`` in :meth:`msdss_data_api.managers.DataManager.get`.
        *args, **kwargs
            Additional arguments passed to :meth:`msdss_data_api.managers.DataManager.get`.

        Author
        ------
        Richard Wen <rrwen.dev@gmail.com>
        
        Example
        -------
        .. code::

            import tempfile
            from celery import Celery
            from msdss_base_database import Database
            from msdss_data_api.managers import DataManager
            from msdss_models_api.models import Model
            from msdss_models_api.managers import *
            from pprint import pprint

            with tempfile.TemporaryDirectory() as folder_path:

                # Setup available models
                models = [Model]
                
                # Create manager
                database = Database()
                data_manager = DataManager(database=database)
                models_manager = ModelsDBManager(models, data_manager, folder=folder_path)

                # Create model instance
                models_manager.create('temp_model', 'Model')

                # Create background manager
                worker = Celery(broker='redis://localhost:6379/0', backend='redis://localhost:6379/0') # rabbitmq
                bg_manager = ModelsDBBackgroundManager(worker, models_manager)

                # Add training data to database
                train_data = [
                    {'col_a': 1, 'col_b': 'a'},
                    {'col_a': 2, 'col_b': 'b'}
                ]
                database.insert('models_test', train_data)

                # Initialize a model instance with inputs from database
                bg_manager.input_db('temp_model', 'models_test')
                
                # Update model instance with new data
                new_data = [
                    {'col_a': 3, 'col_b': 'c'},
                    {'col_a': 4, 'col_b': 'd'}
                ]
                database.insert('models_test', new_data)
                bg_manager.update_db('temp_model', 'models_test', where=['col_a > 2'])
        """
        self.models_manager._load_base_files()

        # (ModelsDBBackgroundManager_update_db_data) Handle database read and query
        where = [split(w) for w in where] if where else where
        self.data_handler.handle_read(dataset)
        self.data_handler.handle_where(where)

        # (ModelsDBBackgroundManager_update_db_handle) Handle model update and processing
        self.models_handler.handle_update(name, self.models_manager.instances)
        self.handler.handle_processing(name, self.states)

        # (ModelsDBBackgroundManager_update_db_add) Add update db task
        self._add_task('update_db', name, dataset, parameters=parameters, where=where, *args, **kwargs)

class ModelsMetadataManager(MetadataManager):
    """
    Class to manage models metadata in a database.

    * Inherits from :class:`msdss_data_api:msdss_data_api.managers.MetadataManager`
    
    Parameters
    ----------
    data_manager : :class:`msdss_data_api.managers.DataManager`
        Data manager object for managing data in a database.
        The restricted tables for the handler will be set to ``[]`` while the only permitted table will be the table name of the parameter ``table``.
    table : str
        The name of the table to store the metadata.
    columns : list(dict) or list(list)
        List of dict (kwargs) or lists (positional args) that are passed to sqlalchemy.schema.Column. See parameter ``columns`` in :meth:`msdss_base_database:msdss_base_database.core.create_table`.
        This defines the table to store the metadata, where the default is:

        .. jupyter-execute::
            :hide-code:

            from msdss_models_api.defaults import *
            from pprint import pprint

            pprint(DEFAULT_METADATA_COLUMNS)

    name_column : str
        Name of the column identifying each entry.
    updated_column : str
        Name of the column for storing the last updated date/time.
    *args, **kwargs
        Additional arguments passed to :class:`msdss_data_api:msdss_data_api.managers.MetadataManager`

    Attributes
    ----------
    See attributes in :class:`msdss_data_api:msdss_data_api.managers.MetadataManager`.
    
    Author
    ------
    Richard Wen <rrwen.dev@gmail.com>
    
    Example
    -------
    .. jupyter-execute::

        from datetime import datetime
        from msdss_base_database import Database
        from msdss_data_api.managers import DataManager
        from msdss_models_api.managers import *
        from msdss_models_api.defaults import *
        
        # Setup database
        db = Database()

        # Check if the metadata table exists and drop if it does
        if db.has_table(DEFAULT_METADATA_TABLE):
            db.drop_table(DEFAULT_METADATA_TABLE)

        # Setup metadata manager
        data_manager = DataManager(database=db)
        mdm = ModelsMetadataManager(data_manager)

        # Add metadata
        metadata = [{
            'title': 'Testing Data',
            'description': 'Data used for testing',
            'tags': 'test exp auto',
            'source': 'Automatically generated from Python',
            'created_by': 'msdss',
            'created_at': datetime.now(),
            'updated_at': datetime.now()
        }]
        mdm.create('test_model', metadata)

        # Get metadata
        metadata_get = mdm.get('test_model')
        
        # Search metadata
        search_results = mdm.search(where=['title = "Testing Data"'])

        # Update metadata
        mdm.update('test_model', {'description': 'NEW DESCRIPTION'})

        # Delete metadata
        mdm.delete('test_model')
    """
    def __init__(
        self,
        data_manager=DataManager(),
        table=DEFAULT_METADATA_TABLE,
        columns=DEFAULT_METADATA_COLUMNS,
        name_column=DEFAULT_NAME_COLUMN,
        updated_column=DEFAULT_UPDATE_COLUMN,
        *args, **kwargs
    ):
        super().__init__(
            data_manager=data_manager,
            table=table,
            columns=columns,
            name_column=name_column,
            updated_column=updated_column,
            *args, **kwargs
        )



import os

from fastapi import HTTPException

class ModelsHandler:
    """
    Class to handle model operations.
    
    Author
    ------
    Richard Wen <rrwen.dev@gmail.com>
    
    Example
    -------
    .. jupyter-execute::

        import tempfile
        from msdss_models_api.models import Model
        from msdss_models_api.managers import ModelsManager
        from msdss_models_api.handlers import ModelsHandler

        with tempfile.TemporaryDirectory() as folder_path:

            # Setup available models
            models = [Model]

            # Create handler
            handler = ModelsHandler()
            
            # Create manager
            models_manager = ModelsManager(models, folder=folder_path, handler=handler)

            # Check if model and instance name available
            # Should not raise exceptions
            handler.handle_create('temp_model', 'Model', models_manager.instances, models_manager.models)

            # Check if instance name available
            # Should not raise exceptions
            handler.handle_write('temp_model', models_manager.instances)

            # Create model instance
            models_manager.create('temp_model', 'Model')

            # Check if instance exists to read
            # Should not raise exceptions
            handler.handle_read('temp_model', models_manager.instances)
    """
    
    def handle_create(self, name, model, instances, models):
        """
        Handle model creation operation.

        Parameters
        ----------
        name : str
            Name of the model instance.
        model : str
            Name of the model to create.
        models : dict(:class:`msdss_models_api.models.Model`)
            Dictionary of available models from parameter ``models``, where:

            * Each key is the class name
            * Each value is the class itself

        instances : dict(:class:`msdss_models_api.models.Model`)
            Dictionary of loaded model instances.
        
        Author
        ------
        Richard Wen <rrwen.dev@gmail.com>
        
        Example
        -------
        .. jupyter-execute::

            import tempfile
            from msdss_models_api.models import Model
            from msdss_models_api.managers import ModelsManager
            from msdss_models_api.handlers import ModelsHandler

            with tempfile.TemporaryDirectory() as folder_path:

                # Setup available models
                models = [Model]

                # Create handler
                handler = ModelsHandler()
                
                # Create manager
                models_manager = ModelsManager(models, folder=folder_path, handler=handler)

                # Check if model and instance name available
                # Should not raise exceptions
                handler.handle_create('temp_model', 'Model', models_manager.instances, models_manager.models)
        """
        if model not in models:
            raise HTTPException(status_code=404, detail='Model not found')
        self.handle_write(name, instances)

    def handle_input(self, name, instances):
        """
        Handle model input operation.

        Parameters
        ----------
        name : str
            Name of the model instance.
        instances : dict(:class:`msdss_models_api.models.Model`)
            Dictionary of loaded model instances.
        
        Author
        ------
        Richard Wen <rrwen.dev@gmail.com>
        
        Example
        -------
        .. jupyter-execute::

            import tempfile
            from msdss_models_api.models import Model
            from msdss_models_api.managers import ModelsManager
            from msdss_models_api.handlers import ModelsHandler

            with tempfile.TemporaryDirectory() as folder_path:

                # Setup available models
                models = [Model]

                # Create handler
                handler = ModelsHandler()
                
                # Create manager
                models_manager = ModelsManager(models, folder=folder_path, handler=handler)

                # Create model instance
                models_manager.create('temp_model', 'Model')

                # Check if instance exists to read
                # Should not raise exceptions
                handler.handle_input('temp_model', models_manager.instances)
        """
        self.handle_read(name, instances)
        instance = instances[name]
        if not instance.metadata['can_input']:
            raise HTTPException(status_code=403, detail='Model instance does not accept inputs')

    def handle_output(self, name, instances):
        """
        Handle model output operation.

        Parameters
        ----------
        name : str
            Name of the model instance.
        instances : dict(:class:`msdss_models_api.models.Model`)
            Dictionary of loaded model instances.
        
        Author
        ------
        Richard Wen <rrwen.dev@gmail.com>
        
        Example
        -------
        .. jupyter-execute::

            import tempfile
            from msdss_models_api.models import Model
            from msdss_models_api.managers import ModelsManager
            from msdss_models_api.handlers import ModelsHandler

            with tempfile.TemporaryDirectory() as folder_path:

                # Setup available models
                models = [Model]

                # Create handler
                handler = ModelsHandler()
                
                # Create manager
                models_manager = ModelsManager(models, folder=folder_path, handler=handler)

                # Create model instance
                models_manager.create('temp_model', 'Model')

                # Check if instance exists to read
                # Should not raise exceptions
                handler.handle_output('temp_model', models_manager.instances)
        """
        self.handle_read(name, instances)
        instance = instances[name]
        if not instance.metadata['can_output']:
            raise HTTPException(status_code=403, detail='Model instance does not produce outputs')
    
    def handle_read(self, name, instances):
        """
        Handle model read operation.

        Parameters
        ----------
        name : str
            Name of the model instance.
        instances : dict(:class:`msdss_models_api.models.Model`)
            Dictionary of loaded model instances.
        
        Author
        ------
        Richard Wen <rrwen.dev@gmail.com>
        
        Example
        -------
        .. jupyter-execute::

            import tempfile
            from msdss_models_api.models import Model
            from msdss_models_api.managers import ModelsManager
            from msdss_models_api.handlers import ModelsHandler

            with tempfile.TemporaryDirectory() as folder_path:

                # Setup available models
                models = [Model]

                # Create handler
                handler = ModelsHandler()
                
                # Create manager
                models_manager = ModelsManager(models, folder=folder_path, handler=handler)

                # Create model instance
                models_manager.create('temp_model', 'Model')

                # Check if instance exists to read
                # Should not raise exceptions
                handler.handle_read('temp_model', models_manager.instances)
        """
        if name not in instances:
            raise HTTPException(status_code=404, detail='Model instance not found')

    def handle_update(self, name, instances):
        """
        Handle model update operation.

        Parameters
        ----------
        name : str
            Name of the model instance.
        instances : dict(:class:`msdss_models_api.models.Model`)
            Dictionary of loaded model instances.
        
        Author
        ------
        Richard Wen <rrwen.dev@gmail.com>
        
        Example
        -------
        .. jupyter-execute::

            import tempfile
            from msdss_models_api.models import Model
            from msdss_models_api.managers import ModelsManager
            from msdss_models_api.handlers import ModelsHandler

            with tempfile.TemporaryDirectory() as folder_path:

                # Setup available models
                models = [Model]

                # Create handler
                handler = ModelsHandler()
                
                # Create manager
                models_manager = ModelsManager(models, folder=folder_path, handler=handler)

                # Create model instance
                models_manager.create('temp_model', 'Model')

                # Check if instance exists to read
                # Should not raise exceptions
                handler.handle_update('temp_model', models_manager.instances)
        """
        self.handle_read(name, instances)
        instance = instances[name]
        if not instance.metadata['can_update']:
            raise HTTPException(status_code=403, detail='Model instance does not allow updates')

    def handle_write(self, name, instances):
        """
        Handle model write operation.

        Parameters
        ----------
        name : str
            Name of the model instance.
        instances : dict(:class:`msdss_models_api.models.Model`)
            Dictionary of loaded model instances.
        
        Author
        ------
        Richard Wen <rrwen.dev@gmail.com>
        
        Example
        -------
        .. jupyter-execute::

            import tempfile
            from msdss_models_api.models import Model
            from msdss_models_api.managers import ModelsManager
            from msdss_models_api.handlers import ModelsHandler

            with tempfile.TemporaryDirectory() as folder_path:

                # Setup available models
                models = [Model]

                # Create handler
                handler = ModelsHandler()
                
                # Create manager
                models_manager = ModelsManager(models, folder=folder_path, handler=handler)

                # Check if instance name available
                # Should not raise exceptions
                handler.handle_write('temp_model', models_manager.instances)
        """
        if name in instances:
            raise HTTPException(status_code=400, detail='Model instance already exists')

class ModelsBackgroundHandler:
    """
    Class to handle model background operations.
    
    Author
    ------
    Richard Wen <rrwen.dev@gmail.com>
    
    Example
    -------
    .. code::

        import tempfile
        from celery import Celery
        from msdss_models_api.models import Model
        from msdss_models_api.managers import *
        from msdss_models_api.handlers import *

        with tempfile.TemporaryDirectory() as folder_path:

            # Setup available models
            models = [Model]

            # Create handler
            handler = ModelsBackgroundHandler()
            
            # Create manager
            models_manager = ModelsManager(models, folder=folder_path, handler=handler)

            # Create bg manager
            worker = Celery(broker='redis://localhost:6379/0', backend='redis://localhost:6379/0') # rabbitmq
            bg_manager = ModelsBackgroundManager(worker, models_manager)

            # Create model instance
            models_manager.create('temp_model', 'Model')

            # Initialize a model instance with inputs as a background process
            train_data = [
                {'col_a': 1, 'col_b': 'a'},
                {'col_a': 2, 'col_b': 'b'}
            ]
            bg_manager.input('temp_model', train_data)

            # Check if instance is processing
            # Should not raise exceptions
            handler.handle_processing('temp_model', bg_manager.states)

            # Check if state exists
            # Should not raise exceptions
            handler.handle_read_state('temp_model', bg_manager.states)
    """

    def handle_cancel(self, name, states):
        """
        Handle model background task cancellation.

        Parameters
        ----------
        name : str
            Name of the model instance.
        states : dict
            Dictionary of processing states for each instance, consisting of the following keys:

            * ``task`` (str): the action that the process is performing - one of: ``INPUT``, ``UPDATE``, ``LOAD``
            * ``result`` (:class:`celery:celery.result.AsyncResult`): celery async object for getting states, ids, etc (see `celery.result <https://docs.celeryproject.org/en/stable/reference/celery.result.html#celery.result.AsyncResult>`_)
            * ``started_at`` (:class:`datetime.datetime`): datetime object for when the task was started
        
        Author
        ------
        Richard Wen <rrwen.dev@gmail.com>
        
        Example
        -------
        .. code::

            import tempfile
            from celery import Celery
            from msdss_models_api.models import Model
            from msdss_models_api.managers import *
            from msdss_models_api.handlers import *

            with tempfile.TemporaryDirectory() as folder_path:

                # Setup available models
                models = [Model]

                # Create handler
                handler = ModelsBackgroundHandler()
                
                # Create manager
                models_manager = ModelsManager(models, folder=folder_path, handler=handler)

                # Create bg manager
                worker = Celery(broker='redis://localhost:6379/0', backend='redis://localhost:6379/0') # rabbitmq
                bg_manager = ModelsBackgroundManager(worker, models_manager)

                # Create model instance
                models_manager.create('temp_model', 'Model')

                # Initialize a model instance with inputs as a background process
                train_data = [
                    {'col_a': 1, 'col_b': 'a'},
                    {'col_a': 2, 'col_b': 'b'}
                ]
                bg_manager.input('temp_model', train_data)

                # Check if instance processing can be cancelled
                # Should raise an exception
                handler.handle_cancel('temp_model', bg_manager.states)
        """
        self.handle_read_state(name, states)
        status = states[name]['result'].state
        if status not in ['PENDING', 'STARTED', 'RETRY']:
            raise HTTPException(status_code=400, detail='Model instance is not processing')

    def handle_processing(self, name, states):
        """
        Handle model background task processing.

        Parameters
        ----------
        name : str
            Name of the model instance.
        states : dict
            Dictionary of processing states for each instance, consisting of the following keys:

            * ``task`` (str): the action that the process is performing - one of: ``INPUT``, ``UPDATE``, ``LOAD``
            * ``result`` (:class:`celery:celery.result.AsyncResult`): celery async object for getting states, ids, etc (see `celery.result <https://docs.celeryproject.org/en/stable/reference/celery.result.html#celery.result.AsyncResult>`_)
            * ``started_at`` (:class:`datetime.datetime`): datetime object for when the task was started
        
        Author
        ------
        Richard Wen <rrwen.dev@gmail.com>
        
        Example
        -------
        .. code::

            import tempfile
            from celery import Celery
            from msdss_models_api.models import Model
            from msdss_models_api.managers import *
            from msdss_models_api.handlers import *

            with tempfile.TemporaryDirectory() as folder_path:

                # Setup available models
                models = [Model]

                # Create handler
                handler = ModelsBackgroundHandler()
                
                # Create manager
                models_manager = ModelsManager(models, folder=folder_path, handler=handler)

                # Create bg manager
                worker = Celery(broker='redis://localhost:6379/0', backend='redis://localhost:6379/0') # rabbitmq
                bg_manager = ModelsBackgroundManager(worker, models_manager)

                # Create model instance
                models_manager.create('temp_model', 'Model')

                # Initialize a model instance with inputs as a background process
                train_data = [
                    {'col_a': 1, 'col_b': 'a'},
                    {'col_a': 2, 'col_b': 'b'}
                ]
                bg_manager.input('temp_model', train_data)

                # Check if instance is processing
                # Should not raise exceptions
                handler.handle_processing('temp_model', bg_manager.states)
        """
        has_status = name in states
        if has_status:
            status = states[name]['result'].state
            if status in ['PENDING', 'STARTED', 'RETRY']:
                raise HTTPException(status_code=400, detail='Model instance still processing')


import os
import pickle

from datetime import datetime
from pydantic import BaseModel
from typing import Optional

class Model:
    """
    Template class to standardize modelling.

    * Methods ``delete``, ``load``, and ``save`` are handled by default using :class:`pickle` and do not need to be defined if there is no need for custom model saving and loading.
    * Methods ``input``, ``output`` and ``update`` need to be defined as they are placeholders for standardized functions of the model
    
    Parameters
    ----------
    file_path : str
        Path to save, load, and delete the model for persistence without the extension. Can be used in methods as ``self.file``.
    file_ext : str
        File extension to save the model in.
    can_input : bool
        Whether the method ``.input`` is defined and available. This is useful for controlling route requests in an API.
    can_output : bool
        Whether the method ``.output`` is defined and available. This is useful for controlling route requests in an API.
    can_update : bool
        Whether the method ``.update`` is defined and available. This is useful for controlling route requests in an API.

    Attributes
    ----------
    instance : obj or None
        An instance of the model initialized with method :meth:`msdss_models_api.models.Model.input`. Initial value is ``None``.
    file : str
        File path to save the model for persistence in. Includes the ``file_ext``.
    last_loaded : :class:`datetime.datetime` or None
        The date and time that the model was last loaded. If ``None``, model has not been loaded. Useful for lazy loading.
    metadata : dict
        Single-value metadata for the model based on the user inputs to provide additional info:

        * ``can_input`` (bool): same as parameter ``can_input``
        * ``can_output`` (bool): same as parameter ``can_output``
        * ``can_update`` (bool): same as parameter ``can_update``

    Author
    ------
    Richard Wen <rrwen.dev@gmail.com>
    
    Example
    -------
    .. jupyter-execute::

        import tempfile
        from msdss_models_api.models import Model

        with tempfile.NamedTemporaryFile() as file:

            # Create template model
            # Note - typically, you want to set all of the following:
            # Model(model=obj, file_path='path/to/model')
            model_path = file.name
            blank_model = Model(file_path=model_path)
            
            # Not saved yet, should be False
            can_load_before = blank_model.can_load()

            # Save the model
            blank_model.save()

            # Saved, should be True
            can_load_after = blank_model.can_load()
            blank_model.load() # loading works after save

            # To use the model, .input() and .output() needs to be defined
            # To update the model, .update() needs to be defined
            # The template model does nothing for these methods
            # It is recommended to extend this model and redefine these methods
            data = [1,2,3,4,5]
            blank_model.input(data)
            blank_model.output(data)
            blank_model.update(data)

            # After saving, you can delete the saved model
            # This will also set .instance to None
            blank_model.delete()
    """
    def __init__(
        self,
        file_path='./model',
        file_ext='pickle',
        can_input=True,
        can_output=True,
        can_update=True
    ):
        self.instance = None
        self.file = f'{file_path}.{file_ext}'
        self.last_loaded = None
        self.metadata = {
            'name': self.__class__.__name__,
            'can_input': can_input,
            'can_output': can_output,
            'can_update': can_update
        }

    def can_load(self):
        """
        Checks if a model can be loaded using the save file.

        Returns
        -------
        bool
            Whether the model can be loaded from the save file or not.
        
        Author
        ------
        Richard Wen <rrwen.dev@gmail.com>
        
        Example
        -------
        .. jupyter-execute::

            import tempfile
            from msdss_models_api.models import Model

            with tempfile.NamedTemporaryFile() as file:

                # Create template model
                model_path = file.name
                blank_model = Model(file_path=model_path)

                # Check if model can be loaded
                blank_model.can_load()
        """
        out = os.path.isfile(self.file)
        return out

    def delete(self, force=False):
        """
        Deletes the saved model and sets the attribute ``.instance`` to ``None``.

        Parameters
        ----------
        force : bool
            Whether to force deletion regardless if the files exist or not.
        
        Author
        ------
        Richard Wen <rrwen.dev@gmail.com>
        
        Example
        -------
        .. jupyter-execute::

            import tempfile
            from msdss_models_api.models import Model

            with tempfile.NamedTemporaryFile() as file:

                # Create template model
                model_path = file.name
                blank_model = Model(file_path=model_path)

                # Save model
                blank_model.save()

                # Delete saved model
                blank_model.delete()
        """
        self.instance = None
        if self.can_load() or force: 
            os.remove(self.file)

    def input(self, data):
        """
        Template method for input data to initialize model.
        
        Requirements:

        * The first argument should be the input data seen in the parameters
        * Other arguments can be defined as any for the model after the first argument
        * Should set ``self.instance`` to the initialized model

        Notes:

        * Does nothing but act as a template reference for class extension
        * This method should be re-defined using a class extension

        Parameters
        ----------
        data : dict or list or :class:`pandas:pandas.DataFrame`
            Data to use for initializing the model. Should accept a ``list`` or ``dict`` to be input in a :class:`pandas:pandas.DataFrame` or the dataframe itself.
        
        Author
        ------
        Richard Wen <rrwen.dev@gmail.com>
        
        Example
        -------
        .. jupyter-execute::

            import tempfile
            from msdss_models_api.models import Model

            with tempfile.NamedTemporaryFile() as file:

                # Create template model
                model_path = file.name
                blank_model = Model(file_path=model_path)

                # Calling this should initialize the model instance
                # blank_model.instance should be set
                train_data = [
                    {'col_a': 1, 'col_b': 'a'},
                    {'col_a': 2, 'col_b': 'b'}
                ]
                blank_model.input(train_data)
        """
        pass

    def load(self, force=False):
        """
        Loads a saved model.

        If the model file has not been changed, skips loading based on ``last_loaded``.

        Parameters
        ----------
        force : bool
            Whether to force loading whether the saved model has been changed or not.
        
        Author
        ------
        Richard Wen <rrwen.dev@gmail.com>
        
        Example
        -------
        .. jupyter-execute::

            import tempfile
            from msdss_models_api.models import Model

            with tempfile.NamedTemporaryFile() as file:

                # Create template model
                model_path = file.name
                blank_model = Model(file_path=model_path)

                # Save model
                blank_model.save()

                # Load model
                blank_model.load()
        """
        if force or self.needs_load():
            with open(self.file, 'rb') as file:
                self.instance = pickle.load(file)
                self.last_loaded = datetime.now()

    def needs_load(self):
        """
        Check if model needs to be loaded again.

        Returns
        -------
        bool
            Whether the model needs to be loaded again based on whether the save ``file`` has changed.
        
        Author
        ------
        Richard Wen <rrwen.dev@gmail.com>
        
        Example
        -------
        .. jupyter-execute::

            import tempfile
            from msdss_models_api.models import Model

            with tempfile.NamedTemporaryFile() as file:

                # Create template model
                model_path = file.name
                blank_model = Model(file_path=model_path)

                # Save model
                blank_model.save()

                # Check if needs load
                blank_model.needs_load()
        """
        last_modified = datetime.fromtimestamp(os.path.getmtime(self.file))
        out = last_modified > self.last_loaded if self.last_loaded else True
        return out

    def output(self, data):
        """
        Template method for a model to output data such as predictions or clusters.
        
        Requirements:

        * The first argument should be the input data seen in the parameters
        * Other arguments can be defined as any for the model output after the first argument
        * Ideally, should use ``self.instance`` to produce the output

        Notes:

        * Does nothing but act as a template reference for class extension
        * This method should be re-defined using a class extension

        Parameters
        ----------
        data : dict or list or :class:`pandas:pandas.DataFrame`
            Data to use as input for the model. Should accept a ``list`` or ``dict`` to be input in a :class:`pandas:pandas.DataFrame` or the dataframe itself.

        Returns
        -------
        :class:`pandas:pandas.DataFrame`
            Output data from the model using the input data from parameter ``data``.
        
        Author
        ------
        Richard Wen <rrwen.dev@gmail.com>
        
        Example
        -------
        .. jupyter-execute::

            import tempfile
            from msdss_models_api.models import Model

            with tempfile.NamedTemporaryFile() as file:

                # Create template model
                model_path = file.name
                blank_model = Model(file_path=model_path)

                # Calling this should initialize the model instance
                # blank_model.instance should be set
                train_data = [
                    {'col_a': 1, 'col_b': 'a'},
                    {'col_a': 2, 'col_b': 'b'}
                ]
                blank_model.input(train_data)

                # Calling this will produce model outputs but only after .input() is used
                # blank_model.instance should be used to produce the outputs
                test_data = [
                    {'col_a': 2, 'col_b': 'c'},
                    {'col_a': 3, 'col_b': 'd'}
                ]
                out = blank_model.output(test_data)
        """
        pass

    def save(self):
        """
        Saves the model to a file to be loaded.
        
        Author
        ------
        Richard Wen <rrwen.dev@gmail.com>
        
        Example
        -------
        .. jupyter-execute::

            import tempfile
            from msdss_models_api.models import Model

            with tempfile.NamedTemporaryFile() as file:

                # Create template model
                model_path = file.name
                blank_model = Model(file_path=model_path)

                # Save model
                blank_model.save()
        """
        with open(self.file, 'wb') as file:
            pickle.dump(self.instance, file)

    def update(self, data):
        """
        Template method for updating a model with new data.
        
        Requirements:

        * The first argument should be the input data seen in the parameters
        * Other arguments can be defined as any for the model output after the first argument
        * Ideally, should update ``self.instance``

        Notes:

        * Does nothing but act as a template reference for class extension
        * This method should be re-defined using a class extension

        Parameters
        ----------
        data : dict or list or :class:`pandas:pandas.DataFrame`
            Data to use for updating the model. Should accept a ``list`` or ``dict`` to be input in a :class:`pandas:pandas.DataFrame` or the dataframe itself.
        
        Author
        ------
        Richard Wen <rrwen.dev@gmail.com>
        
        Example
        -------
        .. jupyter-execute::

            import tempfile
            from msdss_models_api.models import Model

            with tempfile.NamedTemporaryFile() as file:

                # Create template model
                model_path = file.name
                blank_model = Model(file_path=model_path)

                # Calling this should initialize the model instance
                # blank_model.instance should be set
                train_data = [
                    {'col_a': 1, 'col_b': 'a'},
                    {'col_a': 2, 'col_b': 'b'}
                ]
                blank_model.input(train_data)

                # Calling this will update the model but only after .input() is used
                # blank_model.instance should be updated with the new data
                new_data = [
                    {'col_a': 2, 'col_b': 'c'},
                    {'col_a': 3, 'col_b': 'd'}
                ]
                blank_model.update(new_data)
        """
        pass

class ModelCreate(BaseModel):
    """
    Class for creating creating models using the API.
    
    Attributes
    ----------
    title : str
        Title of the model stored in metadata.
    description : str
        Description of model stored in metadata.
    source : str
        Source of model stored in metadata.
    tags : str
        Space separated tags for the model.
    
    Author
    ------
    Richard Wen <rrwen.dev@gmail.com>
    
    Example
    -------
    .. jupyter-execute::

        from msdss_models_api.models import *
        from pprint import pprint
        fields = ModelCreate.__fields__
        pprint(fields)
    """
    title: Optional[str]
    description: Optional[str]
    source: Optional[str]
    tags: Optional[str]

class ModelMetadataUpdate(BaseModel):
    """
    Model for updating model metadata.
    
    Attributes
    ----------
    title : str
        Title of the model stored in metadata.
    description : str
        Description of model stored in metadata.
    source : str
        Data source of model stored in metadata.
    tags : str
        Space separated tags for the model.
    
    Author
    ------
    Richard Wen <rrwen.dev@gmail.com>
    
    Example
    -------
    .. jupyter-execute::

        from msdss_models_api.models import *
        from pprint import pprint
        fields = ModelMetadataUpdate.__fields__
        pprint(fields)
    """
    title: Optional[str]
    description: Optional[str]
    source: Optional[str]
    tags: Optional[str]